import re

MIN_PROFILE_COVERAGE = 75
MIN_MATCH_COVERAGE = 30
NOT_AVAILABLE = 'NA'
NO_CAT  = 'noCat'
NOT_CHIMERIC = 'ok'
INCOMPLETE = 'incomplete'
COMPLETE = 'complete'
POTENTIAL_CHIMERIC = 'PotentialChimeric'
CLASS_I = 'I'
CLASS_II = 'II'
CLASS_I_SINGLE_CONDITIONS = ('termLTR', 'GAG', 'AP', 'RT', 'RH')
CLASS_II_SINGLE_CONDITIONS = ('Tase', 'RPA', 'HEL', 'ATP', 'CYP', 'POLB')

class DecisionMakerPerConsensus(object):
 
    def __init__(self, env):
        self._environment = env
        self._dOrder2Class = {
                      "LTR": CLASS_I,
                      "LARD": CLASS_I,
                      "TRIM": CLASS_I,
                      "DIRS": CLASS_I,
                      "PLE": CLASS_I,
                      "LINE": CLASS_I,
                      "SINE": CLASS_I,
                      "TIR": CLASS_II,
                      "MITE": CLASS_II,
                      "Crypton": CLASS_II,
                      "Maverick": CLASS_II,
                      "Helitron": CLASS_II,
                      "PotentialHostGene": NOT_AVAILABLE,
                      "SSR": NOT_AVAILABLE,
                      "rDNA": NOT_AVAILABLE
                      }
    
    def decidePerConsensus(self, consensusName, lResults, consensusLength):
        lResultsWithClassif = self._getOnlyClassifierResultInstancesWithClassif(lResults)
        resultHG_Profiles, lResultsNotHG_ProfilesWithClassif = self.separateHG_ProfilesAndOtherAgentResults(lResultsWithClassif)
        lResultsOrderWithClassifWithoutHGProfiles, lResultsClassWithClassif = self.separateClassAndOrderAgentResults(lResultsNotHG_ProfilesWithClassif)
        message, orderClassif, chimeric, maxIC = self.getOrderAndMaxICAndMessageAndChimeric(lResults, lResultsOrderWithClassifWithoutHGProfiles)
        consensusStrand = self.getConsensusStrand(consensusName)
        completeness = NOT_AVAILABLE
        if orderClassif == NO_CAT:
            consensusClass, maxIC, message = self.getClassAndIC(lResultsClassWithClassif, lResults, message)
        elif "|" in orderClassif:
            lOrder = orderClassif.split("|")
            consensusClass = self._dOrder2Class[lOrder[0]]
            for order in lOrder[1:]:
                if consensusClass != self._dOrder2Class[order]:
                    consensusClass = NO_CAT
                    break
        else:    
            consensusClass = self._dOrder2Class[orderClassif]  
            completeness = self.getCompletenessAccordingToClassif(message, orderClassif)
                
        if consensusClass == NO_CAT and orderClassif == NO_CAT:
            if self._isPotentialHostGeneWithHG_Profiles(resultHG_Profiles):
                message, maxIC, orderClassif = self._getClassifAndMaxICAndMessageWithBestResult(lResults, resultHG_Profiles)
                consensusClass = NOT_AVAILABLE
            else:
                message = self.createCompleteMessage(lResults)
        
        d2Return = {
                    'consensusName'   : consensusName,
                    'consensusLength' : consensusLength,
                    'consensusStrand' : consensusStrand, 
                    'chimeric'        : chimeric,
                    'consensusClass'  : consensusClass,
                    'consensusOrder'  : orderClassif,
                    'completeness'    : completeness,
                    'confidenceIndex' : maxIC,
                    'dictOfEvidence'  : message
                    }    
        return d2Return

    def getConsensusStrand(self, consensusName):
        lDemandAndResults = self._environment.getResultsOfOneAgent("Strand_Agent")
        for iDemandAndResult in lDemandAndResults:
            if iDemandAndResult.demand.consensusName == consensusName:
                return iDemandAndResult.result.__str__()
        return NOT_AVAILABLE

    def createCompleteMessage(self, lResults):
        allEvidences = {}
        for currentResult in lResults:
            currentDictOfEvidence = currentResult.message['dictOfEvidence']
            if currentDictOfEvidence.has_key("length"):
                del currentDictOfEvidence["length"]
            for currentKey, currentValue in currentDictOfEvidence.items():
                self._addEachEvidenceInMessage(allEvidences, currentKey, currentValue)
        return allEvidences
    
    #TODO: One decision maker for Order 
    def getOrderAndMaxICAndMessageAndChimeric(self, lResults, lResultsOrderWithClassif):
        orderClassif = NO_CAT
        chimeric = NOT_CHIMERIC
        maxIC = NOT_AVAILABLE
        if lResultsOrderWithClassif != []:
            if len(lResultsOrderWithClassif) == 1:
                message, maxIC, orderClassif = self._getOrderAndMaxICAndMessageInNotChimericCase(lResults, lResultsOrderWithClassif)
            else:
                hasLTRResult = False
                hasLARDorTRIMResult = False
                hasTIRResult = False
                hasMITEResult = False
                hasRDNASSRorHGResult = False
                
                for result in lResultsOrderWithClassif:
                    if result.message.get("classif") == "LTR":
                        hasLTRResult = True
                    if result.message.get("classif") == "LARD" or result.message.get("classif") == "TRIM":
                        hasLARDorTRIMResult = True
                        LARDorTRIMResult = result
                        
                    if result.message.get("classif") == "TIR":
                        hasTIRResult = True
                    if result.message.get("classif") == "MITE":
                        hasMITEResult = True
                        MITEResult = result
                        
                    if result.message.get("classif") == "rDNA" or result.message.get("classif") == "SSR" or result.message.get("classif") == "PotentialHostGene":
                        hasRDNASSRorHGResult = True
                        RDNAorSSRResult = result
                        
                if hasRDNASSRorHGResult:
                    lResultsOrderWithClassif = [RDNAorSSRResult]
                    return self.getOrderAndMaxICAndMessageAndChimeric(lResults, lResultsOrderWithClassif)
                
                elif hasLTRResult and hasLARDorTRIMResult:
                    lResultsOrderWithClassif.remove(LARDorTRIMResult)
                    return self.getOrderAndMaxICAndMessageAndChimeric(lResults, lResultsOrderWithClassif)
                        
                elif hasTIRResult and hasMITEResult:
                    lResultsOrderWithClassif.remove(MITEResult)
                    return self.getOrderAndMaxICAndMessageAndChimeric(lResults, lResultsOrderWithClassif)
                else:
                    chimeric = POTENTIAL_CHIMERIC
                    message, maxIC, orderClassif = self._getOrderAndMaxICAndMessageInChimericCase(lResults, lResultsOrderWithClassif)
        else:
            message = self.createCompleteMessage(lResults)
        return message, orderClassif, chimeric, maxIC
        
    
    #TODO: one decision maker for Class
    #TODO: message is not modified in first if
    def getClassAndIC(self, lResultsClassWithClassif, lResults, message):
        classClassif = NO_CAT
        maxCI = NOT_AVAILABLE
        if len (lResultsClassWithClassif) == 0:
            return classClassif, maxCI, message 
        if len (lResultsClassWithClassif) == 1:
            maxCI, classClassif, message = self._getClassAndMaxICInNotChimericCase(lResultsClassWithClassif, lResults)
        else:
            maxCI = -1
            lResultsWithMaxCI = []
            for currentResult in lResultsClassWithClassif:
                if currentResult.message['confidenceIndex'] > maxCI:
                    maxCI = currentResult.message['confidenceIndex']
            for currentResult in lResultsClassWithClassif:
                if maxCI == currentResult.message['confidenceIndex']:
                    lResultsWithMaxCI.append(currentResult)
            if len(lResultsWithMaxCI) == 1:
                message, maxCI, classClassif = self._getClassifAndMaxICAndMessageWithBestResult(lResults, lResultsWithMaxCI[0])
            else:
                maxCI = NOT_AVAILABLE
                message = self.createCompleteMessage(lResults)
            if classClassif == "ClassI":
                classClassif = CLASS_I
            elif classClassif == "ClassII": 
                classClassif = CLASS_II  
        return classClassif, maxCI, message
    
    def separateClassAndOrderAgentResults(self, lResults):
        lResultsOrder = []
        lResultsClass = []
        for currentResult in lResults:
            if re.search("Class", currentResult.demand.agent):
                lResultsClass.append(currentResult)
            else: lResultsOrder.append(currentResult)
        return lResultsOrder, lResultsClass
    
    def separateHG_ProfilesAndOtherAgentResults(self, lResults):
        resultHG_Profiles = None
        lResultsWithoutHG_Profiles = []
        for currentResult in lResults:
            if currentResult.demand.agent != 'HG_profilesAdaptatorAgent':
                lResultsWithoutHG_Profiles.append(currentResult)
            else: resultHG_Profiles = currentResult
        return resultHG_Profiles, lResultsWithoutHG_Profiles

    #TODO: review completeness.
    #    For example : a LTR without match can be complete, but unknown. To be complete, must have termLTR and the 5 profiles (GAG, AP, INT, RT and RH)
    #    Completeness for LARD, TRIM, MITE ? with match and termR ?
    #    LINE must have polyA tail to be complete ?
    def getCompletenessAccordingToClassif(self, message, orderClassif):
        completeness = NOT_AVAILABLE
        if orderClassif == "LTR" or orderClassif == "TIR":
            completeness = self.getCompletenessAccordingToMatchsAndTRAndProfiles(message, orderClassif)
        if orderClassif == "LINE" or orderClassif == "PLE" or orderClassif == "DIRS" or orderClassif == "Crypton":
            completeness = self._getCompletenessAccordingToMatchsAndLengthAndProfiles(message, ["<700bps"])
        if orderClassif == "Helitron":
            completeness = self._getCompletenessAccordingToMatchsAndLengthAndProfiles(message, ["<700bps", ">700bps"])
        if orderClassif == "Maverick":
            completeness = self._getCompletenessAccordingToMatchsAndLengthAndTRAndProfiles(message, "TIR", ["<700bps", ">700bps", ">4000bps"])
        if orderClassif == "SINE":
            completeness = self._getCompletenessAccordingToMatchsAndPolyAtail(message)
        return completeness

    def getCompletenessAccordingToMatchsAndTRAndProfiles(self, dEvidence, orderClassif):
        if self._areTermRPresent(dEvidence, 'term' + orderClassif) \
            and self._areMatchCoverageGreaterThanMIN_MATCH_COVERAGE(dEvidence) \
            and self._areProfilesCoveragesGreaterThanMIN_PROFILE_COVERAGE(dEvidence):
            return COMPLETE
        return INCOMPLETE
    
    def _getCompletenessAccordingToMatchsAndPolyAtail(self, dEvidence):
        if self._isPolyAtailPresent(dEvidence) \
            and self._areMatchCoverageGreaterThanMIN_MATCH_COVERAGE(dEvidence):
            return COMPLETE
        return INCOMPLETE
    
    def _getCompletenessAccordingToMatchsAndLengthAndProfiles(self, dEvidence, lPrecludingLengthCategory):
        if self._isSufficientLength(dEvidence, lPrecludingLengthCategory) \
            and self._areMatchCoverageGreaterThanMIN_MATCH_COVERAGE(dEvidence) \
            and self._areProfilesCoveragesGreaterThanMIN_PROFILE_COVERAGE(dEvidence):
            return COMPLETE
        return INCOMPLETE
    
    def _getCompletenessAccordingToMatchsAndLengthAndTRAndProfiles(self, dEvidence, orderClassif, lPrecludingLengthCategory):
        if self._isSufficientLength(dEvidence, lPrecludingLengthCategory) \
            and self._areTermRPresent(dEvidence, 'term' + orderClassif) \
            and self._areMatchCoverageGreaterThanMIN_MATCH_COVERAGE(dEvidence) \
            and self._areProfilesCoveragesGreaterThanMIN_PROFILE_COVERAGE(dEvidence):
            return COMPLETE
        return INCOMPLETE
    
    def _areMatchCoverageGreaterThanMIN_MATCH_COVERAGE(self, dEvidence):
        return self._isBestMatchCoverageGreaterThanMIN_MATCH_COVERAGE(dEvidence, "Repbase_tbx") \
            or self._isBestMatchCoverageGreaterThanMIN_MATCH_COVERAGE(dEvidence, "Repbase_bx")

    def _isBestMatchCoverageGreaterThanMIN_MATCH_COVERAGE(self, dEvidence, key):
        if dEvidence.has_key(key):
            maxCoverage = 0
            for iBlasterMatch in dEvidence[key]:
                coverage = float(iBlasterMatch.getCoverage())
                if coverage > maxCoverage:
                    maxCoverage = coverage
            return maxCoverage > MIN_MATCH_COVERAGE
        return False
    
    def _areTermRPresent(self, dEvidence, termR):
        if dEvidence.has_key('TR') :
            for evidence in dEvidence["TR"]:
                if (re.search(termR, str(evidence))) :
                    return True
        return False
    
    def _isPolyAtailPresent(self, dEvidence):
        return dEvidence.has_key('polyAtail') or dEvidence.has_key('SSRtrf')
    
    def _isSufficientLength(self, dEvidence, lPrecludingLengthCategory):
        return dEvidence.has_key('length') and dEvidence['length'] not in lPrecludingLengthCategory

    #TODO: check only profiles concerned by the current classification !!!
    def _areProfilesCoveragesGreaterThanMIN_PROFILE_COVERAGE(self, dEvidence):
        profilesGoodCoverageIfExist = True
        if dEvidence.has_key('te_hmmer'):
            noTypeInOrf = False
            dProfilesResults = dEvidence['te_hmmer']
            if dEvidence.has_key('ORF'):
                lORFResults = dEvidence['ORF']
                lCoverages = self._sumCoverageByTypeInTheSameFrameInORF(dProfilesResults, lORFResults)
                if lCoverages == []:
                    noTypeInOrf = True
            if not dEvidence.has_key('ORF') or noTypeInOrf:
                lCoverages = dProfilesResults.values()
            minCoverage = min(lCoverages)    
            if minCoverage > MIN_PROFILE_COVERAGE:
                profilesGoodCoverageIfExist = True
            else:
                profilesGoodCoverageIfExist = False
        return profilesGoodCoverageIfExist
    
    def _sumCoverageByTypeInTheSameFrameInORF(self, dProfilesResults, lORFResults):
        lNbOccurenceProfilesByORF = self._createDictOfNBprofilesByTypeByORF(lORFResults)
        dProfilesStoredByType = self._storeProfilesByType(dProfilesResults)
        lCoverages = self._createListCoverageWithCoveragesAddedIfAllProfilesForEachTypeAreInTheSameORF(lNbOccurenceProfilesByORF, dProfilesStoredByType)    
        return lCoverages 

    def _createDictOfNBprofilesByTypeByORF(self, lORFResults):
        numORF = 0
        lNbOccurenceProfilesByORF = []
        for oRF in lORFResults:
            lORFInfos = oRF.rsplit(" ")
            dNbProfiles = {}
            profilesInfo = False
            previousInfo = ""
            for info in lORFInfos:
                if profilesInfo == True:
                    if info != previousInfo:
                        dNbProfiles[info] = oRF.count(info)
                    previousInfo = info
                profilesInfo = True
            lNbOccurenceProfilesByORF.append(dNbProfiles)
            numORF += 1
        return lNbOccurenceProfilesByORF 

    def _storeProfilesByType(self, dProfilesResults):
        dProfilesStoredByType = {}
        for currentProfile in dProfilesResults.keys():
            profileType = self._getProfileType(currentProfile)
            if (not dProfilesStoredByType.has_key(profileType)):
                dProfilesStoredByType[profileType] = {currentProfile: dProfilesResults[currentProfile]}
            else:
                dict2Append = dProfilesStoredByType[profileType]
                dict2Append[currentProfile] = dProfilesResults[currentProfile]
                dProfilesStoredByType[profileType] = dict2Append
        return dProfilesStoredByType
            
    def _getProfileType(self, currentProfile):
        tabTmp1 = currentProfile.split("_")
        profileType = tabTmp1[-2]
        return profileType  

    def _createListCoverageWithCoveragesAddedIfAllProfilesForEachTypeAreInTheSameORF(self, lNbOccurenceProfilesByORF, dProfilesStoredByType):
        lCoverages = []
        lProfilesConsidered = []
        for type in dProfilesStoredByType.keys():
            for dNbOccurrenceProfiles in lNbOccurenceProfilesByORF:
                if dNbOccurrenceProfiles.has_key(type):
                    if dNbOccurrenceProfiles[type] == len(dProfilesStoredByType[type]):
                        sumCoverage = sum(dProfilesStoredByType[type].values())
                        lCoverages.append(sumCoverage)
                    else:
                        for coverage in dProfilesStoredByType[type].values():
                            if coverage not in lCoverages:
                                lCoverages.append(coverage)
                    lProfilesConsidered.append(dProfilesStoredByType[type])
        self._extendListOfCoveragesWithprofilesNotInORF(dProfilesStoredByType, lCoverages, lProfilesConsidered)
        return lCoverages

    def _extendListOfCoveragesWithprofilesNotInORF(self, dProfilesStoredByType, lCoverages, lProfilesConsidered):
        for dProfiles in dProfilesStoredByType.values():
            if dProfiles not in lProfilesConsidered:
                lCoverages.extend(dProfiles.values()) 
 
    def _getOnlyClassifierResultInstancesWithClassif(self, lResults):
        lResultsWithClassif = []
        for currentResult in lResults:
            if not re.search("!", currentResult.message['classif']):
                lResultsWithClassif.append(currentResult)
        return lResultsWithClassif

    def _isPotentialHostGeneWithHG_Profiles(self, resultHG_Profiles):
        return resultHG_Profiles != None and resultHG_Profiles.message['classif'] == 'PotentialHostGene'
    
    def _addEachEvidenceInMessage(self, allEvidences, evidenceKey, evidenceValue):
        if allEvidences.has_key(evidenceKey) and evidenceKey == 'TR':
            self._addOnlyMissingTermRInfo(allEvidences, evidenceKey, evidenceValue)
        if not allEvidences.has_key(evidenceKey):
            allEvidences[evidenceKey] = evidenceValue
            
    def _addOnlyMissingTermRInfo(self, allEvidences, evidenceKey, evidenceValue):
        for termRElement in evidenceValue:
            if not termRElement in allEvidences[evidenceKey]:
                allEvidences[evidenceKey].append(termRElement)

    def _getOrderAndMaxICAndMessageInNotChimericCase(self, lResults, lResultsWithClassif):
        bestResult = self._partingBestResultAndOtherResults(lResults, lResultsWithClassif)
        message, maxIC, orderClassif = self._getClassifAndMaxICAndMessageWithBestResult(lResults, bestResult)
        return message, maxIC, orderClassif

    def _getOrderAndMaxICAndMessageInChimericCase(self, lResults, lResultsOrderWithClassif):
        maxCI = -1
        lResultsWithMaxCI = []
        for currentResult in lResultsOrderWithClassif:
            if currentResult.message['confidenceIndex'] > maxCI:
                maxCI = currentResult.message['confidenceIndex']
        for currentResult in lResultsOrderWithClassif:
            if maxCI == currentResult.message['confidenceIndex']:
                lResultsWithMaxCI.append(currentResult)
        if len(lResultsWithMaxCI) == 1:
            message, maxCI, orderClassif = self._getClassifAndMaxICAndMessageWithBestResult(lResults, lResultsWithMaxCI[0])
        else:
            orderClassif = "|".join(dResult.message['classif'] for dResult in lResultsWithMaxCI)
            message = self.createCompleteMessage(lResults)
        return message, maxCI, orderClassif
    
    def _getClassAndMaxICInNotChimericCase(self, lResultsClassWithClassif, lResults):
        bestResult = self._partingBestResultAndOtherResults(lResults, lResultsClassWithClassif)
        message, maxIC, classClassif = self._getClassifAndMaxICAndMessageWithBestResult(lResults, bestResult)
        if classClassif == "ClassI":
            classClassif = CLASS_I
        else: 
            classClassif = CLASS_II        
        return  maxIC, classClassif, message

    def _partingBestResultAndOtherResults(self, lResults, lResultsWithClassif):
        bestResult = lResultsWithClassif[0]
        self._removeAnItemFromAList(lResults, bestResult)
        return bestResult

    def _removeAnItemFromAList(self, lResults, bestResult):
        bestIndex = lResults.index(bestResult)
        del lResults[bestIndex]

    def _getClassifAndMaxICAndMessageWithBestResult(self, lResults, bestResult):
        classif, maxIC = self._fillClassifAndIC(bestResult)
        #TODO: don't need yet a deepcopy, because "length" is not an object for the time being...
        message = bestResult.message['dictOfEvidence'].copy()
        self._addAllEvidencesFromNotBestResultsInOtherField(lResults, message)
        return message, maxIC, classif

    def _fillClassifAndIC(self, result):
        IC = result.message['confidenceIndex']
        classif = result.message['classif']
        return classif, IC

    def _addAllEvidencesFromNotBestResultsInOtherField(self, lResults, bestDictOfEvidence):
        dEvidenceType2EvidenceToAdd = {}
        otherEvidence = 'other'
        for currentResult in lResults:
            currentDictOfEvidence = currentResult.message['dictOfEvidence']
            if currentDictOfEvidence.has_key("length"):
                del currentDictOfEvidence["length"]
            for currentKey, currentValue in currentDictOfEvidence.items():
                self._addOtherInMessage(bestDictOfEvidence, currentKey, currentValue, dEvidenceType2EvidenceToAdd)
        if dEvidenceType2EvidenceToAdd != {}:
            bestDictOfEvidence[otherEvidence] = dEvidenceType2EvidenceToAdd
            
    def _addOtherInMessage(self, bestDictOfEvidence, evidenceKey, evidenceValue, dEvidenceType2EvidenceToAddInFieldOther):
        if evidenceKey == 'TR':
            lTR = self._initializeTRList(dEvidenceType2EvidenceToAddInFieldOther)
            if bestDictOfEvidence.has_key(evidenceKey):
                for termRElement in evidenceValue:
                    if not termRElement in bestDictOfEvidence[evidenceKey]:
                        self._fillTRListIfTermRElementIsNotInDict(dEvidenceType2EvidenceToAddInFieldOther, lTR, termRElement)
            else:
                for termRElement in evidenceValue:
                    self._fillTRListIfTermRElementIsNotInDict(dEvidenceType2EvidenceToAddInFieldOther, lTR, termRElement)
            if lTR != []:
                dEvidenceType2EvidenceToAddInFieldOther[evidenceKey] = lTR
        if evidenceKey == "Repbase_bx" or evidenceKey == "Repbase_bx":
            for iBlasterMatch in evidenceValue:
                if not bestDictOfEvidence.has_key(evidenceKey) or iBlasterMatch not in bestDictOfEvidence[evidenceKey]:
                    if not dEvidenceType2EvidenceToAddInFieldOther.has_key(evidenceKey):
                        dEvidenceType2EvidenceToAddInFieldOther[evidenceKey] = [iBlasterMatch]
                    elif iBlasterMatch not in dEvidenceType2EvidenceToAddInFieldOther[evidenceKey]:
                        dEvidenceType2EvidenceToAddInFieldOther[evidenceKey].append(iBlasterMatch)
        elif not bestDictOfEvidence.has_key(evidenceKey) and not dEvidenceType2EvidenceToAddInFieldOther.has_key(evidenceKey):
            dEvidenceType2EvidenceToAddInFieldOther[evidenceKey] = evidenceValue

    def _fillTRListIfTermRElementIsNotInDict(self, dEvidenceType2EvidenceToAddInFieldOther, lTR, termRElement):
        if not dEvidenceType2EvidenceToAddInFieldOther.has_key('TR'):
            lTR.append(termRElement)
        elif not termRElement in dEvidenceType2EvidenceToAddInFieldOther['TR']:
            lTR.append(termRElement)

    def _initializeTRList(self, dEvidenceType2EvidenceToAddInFieldOther):
        if not dEvidenceType2EvidenceToAddInFieldOther.has_key('TR'):
            lTR = []
        else:
            lTR = dEvidenceType2EvidenceToAddInFieldOther['TR']
        return lTR