#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.checker.ConfigChecker import ConfigRules, ConfigChecker
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory
from commons.core.launcher.Launcher import Launcher
from commons.core.utils.FileUtils import FileUtils
import subprocess
import os

LOG_DEPTH = "repet.tools"

class SelfAlign(object):
    
    def __init__(self, pipelineName, configFileName, verbosity = 1):
        self._pipelineName = pipelineName
        self._configFileName = configFileName
        self._sectionName = "self_align"
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)

    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section=self._sectionName, mandatory=True)
        iConfigRules.addRuleOption(section=self._sectionName, option ="resources", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="tmpDir", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="copy", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="blast", mandatory=True, type="string", set = ("ncbi", "blastplus", "wu"))
        iConfigRules.addRuleOption(section=self._sectionName, option ="Evalue", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="length", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="identity", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="filter_HSP", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="min_Evalue", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="min_identity", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="min_length", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="max_length", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="clean", mandatory=True, type="bool")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._projectName = self._iConfig.get("project", "project_name")
        self._projectDir = self._iConfig.get("project", "project_dir")
        self._chunkLength = self._iConfig.get("prepare_batches", "chunk_length")
        self._chunkOverlap = self._iConfig.get("prepare_batches", "chunk_overlap")
        self._resources = self._iConfig.get(self._sectionName, "resources")
        self._tmpDir = self._iConfig.get(self._sectionName, "tmpDir")
        self._isCopyOnNode = self._iConfig.get(self._sectionName, "copy")
        self._blastToUse = self._iConfig.get(self._sectionName, "blast")
        self._eValue = self._iConfig.get(self._sectionName, "Evalue")
        self._length = self._iConfig.get(self._sectionName, "length")
        self._identity = self._iConfig.get(self._sectionName, "identity")
        self._doFilterHSP = self._iConfig.get(self._sectionName, "filter_HSP")
        self._minEvalue = self._iConfig.get(self._sectionName, "min_Evalue")
        self._minIdentity = self._iConfig.get(self._sectionName, "min_identity")
        self._minLength = self._iConfig.get(self._sectionName, "min_length")
        self._maxLength = self._iConfig.get(self._sectionName, "max_length")
        self._doClean = self._iConfig.get(self._sectionName, "clean")
        
        if self._isCopyOnNode and not self._tmpDir:
            self._isCopyOnNode = False
        self._log.debug("The copy option is: %s." % self._isCopyOnNode)
        
    def _checkAttributes(self):
        if os.path.exists("%s_Blaster" % self._projectName):
            self._logAndRaise("ERROR: directory %s_Blaster already exists" % self._projectName)

        if self._blastToUse == "blastplus":
            if not (CheckerUtils.isExecutableInUserPath("blastn") or CheckerUtils.isExecutableInUserPath("makeblastdb")):
                self._logAndRaise("ERROR: ncbi blast plus (blastn executable) must be in your path")
            else:
                process = subprocess.Popen("blastn -h", shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
                output = process.communicate()[0]
                if "washu" in output.lower():
                    self._logAndRaise("ERROR: you are trying to use NCBI BlastPlus, but the blastn executable is pointing to WuBlast. Please check your system path.")
        elif self._blastToUse == "wu":
            if not (CheckerUtils.isExecutableInUserPath("blastn") or CheckerUtils.isExecutableInUserPath("xdformat")):
                self._logAndRaise("ERROR: WuBlast (blastn executable) must be in your path")
            else:
                process = subprocess.Popen("blastn -h", shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
                output = process.communicate()[0]
                if not "washu" in output.lower():
                    self._logAndRaise("ERROR: you are trying to use WuBlast, but the blastn executable is pointing to NCBI BlastPlus. Please check your system path.")
        else:
            if not (CheckerUtils.isExecutableInUserPath("blastall") or CheckerUtils.isExecutableInUserPath("formatdb")):
                self._logAndRaise("ERROR: ncbi blastall and formatdb must be in your path")

    def _createFilterAlignCmd(self, iLauncher):
        lArgs = []
        lArgs.append("-i %s.align.not_over" % self._projectName)
        lArgs.append("-E %s" % self._minEvalue)
        lArgs.append("-I %s" % self._minIdentity)
        lArgs.append("-l %s" % self._minLength)
        lArgs.append("-L %s" % self._maxLength)
        lArgs.append("-v %d" % (self._verbosity - 1))
        return iLauncher.getSystemCommand("FilterAlign.py", lArgs)  
        
    def run(self):
        self._checkConfig()
        self._checkAttributes()
        self._log.info("START of self-alignment with Blaster")
            
        os.makedirs("%s_Blaster" % self._projectName)
        os.chdir("%s_Blaster" % self._projectName)
        os.symlink("../%s" % self._configFileName, "%s" % self._configFileName)
                                      
        cDir = os.getcwd()
        if not self._tmpDir:
            self._tmpDir = cDir
    
        #TODO: remove temporary config file when TEdenovo.cfg and TEannot.cfg have been merged
        tempConfigfileName = "%s/blaster.cfg" % os.getcwd()
        with open(tempConfigfileName, "w") as fh:
            fh.write("[jobs]\n")
            fh.write("resources: %s\n" % self._resources)
            fh.write("tmpDir: %s\n" % self._tmpDir)
            fh.write("copy: %s\n" % self._isCopyOnNode) 
            fh.write("clean: %s\n" % self._doClean)
            fh.write("\n")
            fh.write("[prepare_data]\n")
            fh.write("chunk_length: %s\n" % self._chunkLength)
            fh.write("chunk_overlap: %s\n" % self._chunkOverlap)
            fh.write("\n")
            fh.write("[alignment]\n")
            fh.write("blast: %s\n" % self._blastToUse)
            fh.write("Evalue: %s\n" % self._eValue)
            fh.write("length: %s\n" % self._length)
            fh.write("identity: %s\n" % self._identity)
            
        batchesDir = "%s/%s_db/batches" % (self._projectDir, self._projectName)
        subjectFilePath = "%s/%s_db/%s_chunks.fa" % (self._projectDir, self._projectName, self._projectName)
        outFileName = "%s.align.not_over" % self._projectName
        groupid = "%s_%s_SelfAlign" % (self._projectName, self._pipelineName)
        cmd = "LaunchBlasterInParallel.py -q %s -s %s -a -o %s -C %s -g %s" % (batchesDir, subjectFilePath, outFileName, tempConfigfileName, groupid)
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        os.remove(tempConfigfileName)
    
        # filter the HSPs
        if self._doFilterHSP:
            groupid = "%s_%s_FilterAlign" % (self._projectName, self._pipelineName)
            acronym = "FilterAlign"
            iDb = DbFactory.createInstance()
            jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, self._tmpDir, "jobs", self._resources, groupid)
            
            lCmdsTuples = []
            lCmds = []    
            lCmds.append(self._createFilterAlignCmd(iLauncher))
            lCmdStart = []
            lCmdStart.append("os.symlink('%s/%s.align.not_over', '%s.align.not_over')" % (cDir, self._projectName, self._projectName))
            lCmdFinish = []
            lCmdFinish.append("shutil.move('%s.align.not_over.filtered', '%s/.')" % (self._projectName, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
            iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)    
        
        if self._doClean:
            if self._doFilterHSP:
                os.remove("%s.align.not_over" % self._projectName)
            os.remove(self._configFileName)
                    
            os.chdir("%s/%s_db" % (self._projectDir, self._projectName))
            FileUtils.removeFilesByPattern("%s*_cut*" % self._projectName)
    
        os.chdir(self._projectDir)
        self._log.info("END SelfAlign")