#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

 
import os
import sys
import getopt
import time
import glob
import ConfigParser
import subprocess
import shutil

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append(os.environ["REPET_PATH"])
if not "PYTHONPATH" in os.environ.keys():
    os.environ["PYTHONPATH"] = os.environ["REPET_PATH"]
else:
    os.environ["PYTHONPATH"] = "%s:%s" % (os.environ["REPET_PATH"], os.environ["PYTHONPATH"])

from ConfigParser import NoOptionError
from ConfigParser import NoSectionError
from ConfigParser import MissingSectionHeaderError
from commons.core.launcher.Launcher import Launcher
from commons.core.checker.RepetException import RepetException
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.checker.CheckerException import CheckerException
from commons.core.utils.FileUtils import FileUtils
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.sql.RepetJob import RepetJob
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory
from commons.tools.MergeMatchsFiles import MergeMatchsFiles

#------------------------------------------------------------------------------

## Setup the required environment.
#
# @param config ConfigParser instance 
#
def setup_env(config):
    os.environ["REPET_HOST"] = config.get("repet_env", "repet_host")
    os.environ["REPET_USER"] = config.get("repet_env", "repet_user")
    os.environ["REPET_PW"] = config.get("repet_env", "repet_pw")
    os.environ["REPET_DB"] = config.get("repet_env", "repet_db")
    os.environ["REPET_PORT"] = config.get("repet_env", "repet_port")
    os.environ["REPET_JOB_MANAGER"] = config.get("repet_env", "repet_job_manager")
    os.environ["REPET_QUEUE"] = config.get("repet_env", "repet_job_manager")
    os.environ["REPET_JOBS"] = "MySQL"

#------------------------------------------------------------------------------

def help():
    print
    print "usage: TEannot.py [ options ]"
    print "options:"
    print "     -h: this help"
    print "     -P: project name (<=15 character only alphanumeric or underscore)"
    print "     -C: configuration file"
    print "     -S: step (1/2 with -a/3 with -c/4 with -s/5/6 with -b/7/8 with -o)"
    print "     -a: program to align sequences (BLR/RM/CEN)"
    print "     -r: use randomized chunks (with \"-S 2\")"
    print "     -c: combine the alignment programs (default=BLR+RM+CEN)"
    print "     -s: program to detect short repeats (TRF/Mreps/RMSSR)"
    print "     -b: blast for step 6 (tblastx/blastx)"
    print "     -o: output format for the annotations (gameXML or GFF3)"
    print "     -v: verbose (0/default=1/2/3)"
    print

#------------------------------------------------------------------------------

def main():

    """
    Pipeline for the annotation of transposable elements (TEs) in genomic sequences.
    """

    global projectName
    projectName = ""
    global configFileName
    configFileName = ""
    step = ""
    align = ""
    useRndChunks = False
    combAlign = "BLR+RM+CEN"
    sat = ""
    typeBlast = ""
    outFormat = ""
    global clean
    clean = False
    global verbose
    verbose = 1

    try:
        opts = getopt.getopt(sys.argv[1:], "hP:C:S:a:rc:s:b:o:v:")[0]
    except getopt.GetoptError, err:
        print str(err)
        help()
        sys.exit(1)
    for o,a in opts:
        if o == "-h":
            help()
            sys.exit(0)
        elif o == "-P":
            projectName = a
        elif o == "-C":
            configFileName = a
        elif o == "-S":
            step = a
        elif o == "-a":
            align = a
        elif o == "-r":
            useRndChunks = True
        elif o == "-c":
            combAlign = a
        elif o == "-s":
            sat = a
        elif o == "-b":
            typeBlast = a
        elif o == "-o":
            outFormat = a
        elif o == "-v":
            verbose = int(a)

    if projectName == "":
        print "ERROR: no project provided (-P)"
        help()
        sys.exit(1)
    if configFileName == "":
        print "ERROR: no configuration file provided (-C)"
        help()
        sys.exit(1)
    if step == "":
        print "ERROR: no step provided (-S)"
        help()
        sys.exit(1)
        
    if not CheckerUtils.isMax15Char(projectName):
        print "ERROR: project name must have 15 character max"
        help()
        sys.exit(1)
        
    if not CheckerUtils.isCharAlphanumOrUnderscore(projectName):
        print "ERROR: project name must contain only alphanumeric or underscore character"
        help()
        sys.exit(1)

    global config
    config = ConfigParser.ConfigParser()
    try:
        config.readfp( open(configFileName) )
    except MissingSectionHeaderError:
        print "ERROR: your config file " + configFileName + " must begin with a section name"
        sys.exit(1)

    isExceptionRaised = False
    sectionName = "repet_env"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_version") 
    except NoOptionError:
        print "ERROR: the option repet_version must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_host") 
    except NoOptionError:
        print "ERROR: the option repet_host must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_user") 
    except NoOptionError:
        print "ERROR: the option repet_user must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_pw") 
    except NoOptionError:
        print "ERROR: the option repet_pw must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_db") 
    except NoOptionError:
        print "ERROR: the option repet_db must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "repet_port") 
    except NoOptionError:
        print "ERROR: the option repet_port must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)

    setup_env(config)

    changeLogFileName = os.environ["REPET_PATH"] + "/annot_pipe/doc/CHANGELOG" 
    changeLogFileHandler = open(changeLogFileName, "r")
    try:
        CheckerUtils.checkConfigVersion(changeLogFileHandler, config)
    except CheckerException, e:
        print e.msg
        sys.exit(1)
    
    isExceptionRaised = False
    sectionName = "project"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "project_name") 
    except NoOptionError:
        print "ERROR: the option project_name must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "project_dir") 
    except NoOptionError:
        print "ERROR: the option project_dir must be defined in " + sectionName +" in your config file : " + configFileName
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)

    if verbose > 0:
        print "\nSTART %s" % (sys.argv[0].split("/")[-1])
        print "version %s" % ( config.get("repet_env","repet_version") )
        now = time.localtime()
        print "local date/time = %s-%s-%s / %s:%s:%s" % ( now[0], now[1], now[2], now[3], now[4], now[5] )
        sys.stdout.flush()

    if config.get( "project", "project_name" ) != projectName:
        print "ERROR: project name different between configuration file ('%s') and command-line option ('%s')" % ( config.get( "project", "project_name" ), projectName )
        sys.exit(1)
    if verbose > 0:
        print "project name = %s" % ( projectName )

    global projectDir
    projectDir = config.get( "project", "project_dir" )
    if projectDir[-1] == "/":
        projectDir = projectDir[:-1]
    if os.path.exists( projectDir ):
        if verbose > 0:
            print "project directory = %s" % ( projectDir ); sys.stdout.flush()
        os.chdir( projectDir )
    else:
        print "ERROR: project directory %s doesn't exist" % ( projectDir )
        sys.exit(1)

    global jobsSectionName
    jobsSectionName = "parallelized_jobs"
    isExceptionRaised = False
    try:
        CheckerUtils.checkSectionInConfigFile(config, jobsSectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (jobsSectionName, configFileName)
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, jobsSectionName, "resources") 
    except NoOptionError:
        print "ERROR: the option 'resources' must be defined in %s in your config file : %s" % (jobsSectionName, configFileName)
        isExceptionRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, jobsSectionName, "tmpDir") 
    except NoOptionError:
        print "ERROR: the option 'tmpDir' must be defined in %s in your config file : %s" % (jobsSectionName, configFileName)
        isExceptionRaised = True
    if isExceptionRaised: 
        sys.exit(1)
    global queue
    queue = config.get(jobsSectionName, "resources")
    if config.get(jobsSectionName, "clean") == "yes":
        clean = True
    
    if step == "1":            
        prepareInputFilesAndLoadTables()

    elif step == "2":
        if align == "":
            print "ERROR: missing option -a"
            help()
            sys.exit(1)
        if useRndChunks == False:
            alignChunksAndTEs( "", align )
        else:
            alignChunksAndTEs( "_rnd", align )

    elif step == "3":
        filterAndCombineTEsAnnot( combAlign )

    elif step == "4":
        if sat == "":
            print "ERROR: missing option -s"
            help()
            sys.exit(1)
        detectSSRs( sat )

    elif step == "5":
        filterAndCombineSSRsAnnot()

    elif step == "6":
        if typeBlast in [ "tblastx", "blastx" ]:
            alignChunksAndTEsForCoding( typeBlast )
        else:
            print "ERROR: blast '%s' not recognized" % ( typeBlast )
            sys.exit(1)
    
    elif step == "7":
        pathsProcessing( config )
        
    elif step == "8":
        if outFormat == "":
            print "ERROR: missing option -o"
            help()
            sys.exit(1)
        if outFormat.lower() in ["gff3", "gamexml"]:
            exportAnnotations( outFormat )
        else:
            print "ERROR: the format '%s' is not recognized" % ( outFormat )
            sys.exit(1)
    
    if verbose > 0:
        now = time.localtime()
        print "local date/time = %s-%s-%s / %s:%s:%s" % ( now[0], now[1], now[2], now[3], now[4], now[5] )
        print "version %s" % config.get("repet_env", "repet_version")
        print "END %s\n" % (sys.argv[0].split("/")[-1])
        sys.stdout.flush()

    return 0

#------------------------------------------------------------------------------

def prepareInputFilesAndLoadTables():
 
    """
    Step 1:
    * cut the input genomic sequences into chunks and load them in a MySQL table
    * randomize the chunks and load them in a MySQL table
    * load the reference TE library in a MySQL table
    * prepare the reference TE library for Blaster (blastn)
    """
    
    if verbose > 0:
        print "\nbeginning of step 1"
        print "prepare the input files and load them in MySQL tables"
        sys.stdout.flush()
    
    genomeFastaFileName = "%s/%s.fa" % ( projectDir, projectName ) 
    if not os.path.exists( genomeFastaFileName ):
        print "ERROR: %s/%s.fa doesn't exist" % ( projectDir, projectName )
        sys.exit(1)
        
    separator = "\n"
    
    inGenomeFileHandler = open(genomeFastaFileName, "r")
    try:
        CheckerUtils.checkHeaders(inGenomeFileHandler)
    except CheckerException, e:
        print "Error in file %s. Wrong headers are :" % genomeFastaFileName
        print separator.join(e.messages)
        print "Authorized characters are : a-z A-Z 0-9 - . : _\n"
        inGenomeFileHandler.close()
        sys.exit(1)
    inGenomeFileHandler.close()
    
    FileUtils.fromWindowsToUnixEof( genomeFastaFileName )
    
    refTEFileName = "%s/%s_refTEs.fa" % ( projectDir, projectName )
    if not os.path.exists( refTEFileName ):
        print "ERROR: %s/%s_refTEs.fa doesn't exist" % ( projectDir, projectName )
        sys.exit(1)

    refTEFileHandler = open(refTEFileName, "r")
    try:
        CheckerUtils.checkHeaders(refTEFileHandler)
    except CheckerException, e:
        print "Error in file %s. Wrong headers are :" % refTEFileName
        print separator.join(e.messages)
        print "Authorized characters are : a-z A-Z 0-9 - . : _\n"
        inGenomeFileHandler.close()
        sys.exit(1)
    refTEFileHandler.close()
    
    FileUtils.fromWindowsToUnixEof( refTEFileName )
        
    if os.path.exists( "%s_db" % ( projectName ) ):
        os.system( "rm -rf %s_db" % ( projectName ) )
    os.mkdir("%s_db" % projectName)
    os.chdir("%s_db" % projectName)

    #--------------------------------------------------------------------------

    if not os.path.exists("%s/%s_db/%s.fa" % (projectDir, projectName, projectName)):
        os.symlink("../%s.fa" % projectName, "%s.fa" % projectName)
    if not os.path.exists("%s/%s_db/%s_refTEs.fa" % (projectDir, projectName, projectName)):
        os.symlink("../%s_refTEs.fa" % projectName, "%s_refTEs.fa_initH" % projectName)
        
    isExceptionRaised = False
    sectionName = "prepare_data"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_length") 
    except NoOptionError:
        print "ERROR: the option 'chunk_length' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "chunk_overlap") 
    except NoOptionError:
        print "ERROR: the option 'chunk_overlap' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "nb_seq_per_batch") 
    except NoOptionError:
        print "ERROR: the option 'nb_seq_per_batch' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "make_random_chunks") 
    except NoOptionError:
        print "ERROR: the option 'make_random_chunks' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    if isExceptionRaised: 
        sys.exit(1)

    # cut the genomic sequences into chunks
    chunkFilePrefix = "%s_chunks" % projectName
    chunkLength = int(config.get(sectionName, "chunk_length"))
    chunkOverlap = int(config.get(sectionName, "chunk_overlap"))
    FastaUtils.dbChunks(genomeFastaFileName, chunkLength, chunkOverlap, 0, chunkFilePrefix, clean, verbose - 2)
    
    # allocate one (or several) chunks(s) into the same file (-> batch)
    prg = "dbSplit.py"
    cmd = prg
    cmd += " -i %s_chunks.fa" % projectName
    cmd += " -n %s" % config.get(sectionName, "nb_seq_per_batch")
    cmd += " -d"
    cmd += " -v %i" % ( verbose - 2 )
    process = subprocess.Popen(cmd, shell = True)
    process.communicate()
    if process.returncode != 0:
        raise Exception("ERROR when launching '%s'" % cmd)
    
    # shuffle each chunk independently to get randomized chunks
    if config.get(sectionName, "make_random_chunks") == "yes":
        if os.path.exists("batches_rnd"):
            os.rmdir("batches_rnd")
        prg = "dbShuffle.py"
        cmd = prg
        cmd += " -I batches"
        cmd += " -O batches_rnd"
        cmd += " -v %i" % ( verbose - 2 )
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)

    # shorten sequence headers
    prg = "ChangeSequenceHeaders.py"
    cmd = prg
    cmd += " -i %s_refTEs.fa_initH" % projectName
    cmd += " -f fasta"
    cmd += " -s 1"
    cmd += " -p refTE_"
    cmd += " -o %s_refTEs.fa" % projectName
    cmd += " -v %i" % ( verbose - 2 )
    process = subprocess.Popen(cmd, shell = True)
    process.communicate()
    if process.returncode != 0:
        raise Exception("ERROR when launching '%s'" % cmd)
   
    iDb = DbFactory.createInstance()
    iDb.createTable("%s_chr_seq" % projectName, "fasta", "%s.fa" % projectName, True)
    iDb.createTable("%s_chk_seq" % projectName, "fasta", "%s_chunks.fa" % projectName, True)
    iDb.createTable("%s_chk_map" % projectName, "map", "%s_chunks.map" % projectName, True)
    iDb.createTable("%s_refTEs_seq" % projectName, "fasta", "%s_refTEs.fa_initH" % projectName, True)
    iDb.createTable("%s_refTEs_map" % projectName, "map", "%s_refTEs.fa_initH.newHlink" % projectName, True)
    iDb.close()

    # prepare the TE library for BLASTER
    sectionName = "align_refTEs_with_genome"
    if config.get(sectionName, "BLR_blast") == "ncbi":
        if not CheckerUtils.isExecutableInUserPath("formatdb"):
            print "ERROR: 'formatdb' from NCBI-BLAST must be in your path"
            sys.exit(1)
    elif config.get(sectionName, "BLR_blast") == "wu":
        if not CheckerUtils.isExecutableInUserPath("xdformat"):
            print "ERROR: 'xdformat' from WU-BLAST must be in your path"
            sys.exit(1)
    elif config.get(sectionName,"BLR_blast") == "blastplus":
        if not CheckerUtils.isExecutableInUserPath("makeblastdb"):
            print "ERROR: 'makeblastdb' from NCBI-BLASTPLUS must be in your path"
            sys.exit(1)  
            
    prg = "blaster"
    cmd = prg
    cmd += " -q %s_refTEs.fa" % ( projectName )
    cmd += " -B %s_refTEs.fa-BLRnPrepare" % ( projectName )
    cmd += " -P"
    if config.get(sectionName, "BLR_blast") == "ncbi": 
        cmd += " -N"
        cmd += " -p '-a 1'"
    elif config.get(sectionName, "BLR_blast") == "wu":
        cmd += " -W"
        cmd += " -p -cpus=1"
    elif config.get(sectionName, "BLR_blast") == "blastplus":
        cmd += " -X"
        cmd += " -p '-num_threads 1'"
    process = subprocess.Popen(cmd, shell = True)
    process.communicate()
    if process.returncode != 0:
        raise Exception("ERROR when launching '%s'" % cmd)

    os.chdir( ".." )

    if verbose > 0:
        print "step 1 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def getBlasterCmd(verbose, projectName, sectionName, config, iLauncher, file):
    lArgs = []
    lArgs.append("-q %s" % file)
    lArgs.append("-s %s_refTEs.fa" % projectName)
    lArgs.append("-B %s" % file)
    lArgs.append("-S %s" % config.get(sectionName, "BLR_sensitivity"))
    if config.get(sectionName, "BLR_blast") == "ncbi": 
        lArgs.append("-N")
        lArgs.append("-p '-a 1'")
    elif config.get(sectionName, "BLR_blast") == "wu":
        lArgs.append("-W")
        lArgs.append("-p -cpus=1")
    elif config.get(sectionName, "BLR_blast") == "blastplus":
        lArgs.append("-X")
        lArgs.append("-p '-num_threads 1'")
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("blaster", lArgs)

def getRepeatMaskerCmd(projectName, sectionName, config, iLauncher, file):
    lArgs = []
    lArgs.append("-cutoff 200")
    if config.get(sectionName, "RM_engine") == "wu": 
        lArgs.append("-e wublast")
    elif config.get(sectionName, "RM_engine") == "cm":
        lArgs.append("-e crossmatch")
    elif config.get(sectionName, "RM_engine") == "ncbi":
        lArgs.append("-e ncbi")
    if config.get(sectionName, "RM_sensitivity") == "s":
        lArgs.append("-s")
    elif config.get(sectionName, "RM_sensitivity") == "q":
        lArgs.append("-q")
    elif config.get(sectionName, "RM_sensitivity") == "qq":
        lArgs.append("-qq")
    lArgs.append("-dir .")
    lArgs.append("-pa 1 -gccalc -no_is -nolow")
    lArgs.append("-lib %s_refTEs.fa" % projectName)
    lArgs.append(file)
    return iLauncher.getSystemCommand("RepeatMasker", lArgs)

def getRM2AlignCmd(iLauncher, file):
    lArgs = []
    lArgs.append("-i %s" % file)
    return iLauncher.getSystemCommand("RMcat2align.py", lArgs)

def getCensorCmd(projectName, sectionName, iLauncher, file):    
    lArgs = []
    lArgs.append(file)
    lArgs.append("-debug")
    lArgs.append("-lib %s_refTEs.fa" % projectName)
    lArgs.append("-s -ns")
    lArgs.append("-nofilter")
    if config.get(sectionName, "CEN_blast") == "ncbi":
        lArgs.append("-bprm '-a 1'")
        return iLauncher.getSystemCommand("censor.ncbi", lArgs)
    elif config.get(sectionName, "CEN_blast") == "wu":
        lArgs.append("-bprm '-cpus 1'")
        return iLauncher.getSystemCommand("censor", lArgs)
    else:
        raise Exception("Censor can't handle %s (change 'BLR_blast' option in configuration file)" % config.get(sectionName, "CEN_blast"))

def getCensor2AlignCmd(iLauncher, file):
    lArgs = []
    lArgs.append("-i %s" % file)
    return iLauncher.getSystemCommand("CENmap2align.py", lArgs)
   
#------------------------------------------------------------------------------

def alignChunksAndTEs( qryDirSuffix, align ):

    """
    Step 2: map the sequence from the TE library on the genomic chunks

    @param qryDirSuffix: query directory suffix ('' or '_rnd')
    @type qryDirSuffix: string

    @param align: alignment program (Blaster, RepeatMasker or Censor)
    @type align: string
    """

    if verbose > 0:
        print "\nbeginning of step 2"
        if qryDirSuffix == "":
            print "use %s to align the TE sequences on the chunks" % ( align )
        else:
            print "use %s to align the TE sequences on the randomized chunks" % ( align )
            if not os.path.exists( projectDir + "/%s_db/batches_rnd" % ( projectName ) ):
                print "ERROR: you didn't build random chunks at step 1, check the configuration file"
                sys.exit(1)
        sys.stdout.flush()

    if not os.path.exists( "%s_TEdetect%s" % ( projectName, qryDirSuffix ) ):
        os.mkdir( "%s_TEdetect%s" % ( projectName, qryDirSuffix ) )
    os.chdir( "%s_TEdetect%s" % ( projectName, qryDirSuffix ) )
    if not os.path.exists( align ):
        os.mkdir( align )
    else:
        print "ERROR: directory %s_TEdetect%s/%s/ already exists" % ( projectName, qryDirSuffix, align )
        sys.exit(1)
    os.chdir( align )

    # launch the alignment program (BLR+M or RM or CEN) on all the chunks in parallel
    if align == "RM":
        if not CheckerUtils.isExecutableInUserPath("RepeatMasker"):
            print "ERROR: RepeatMasker must be in your path"
            sys.exit(1)
    if align == "CEN":
        if not CheckerUtils.isExecutableInUserPath("censor"):
            print "ERROR: censor must be in your path"
            sys.exit(1)
        
    sectionName = "parallelized_jobs"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "copy")
    except NoOptionError:
        print "ERROR: the option 'copy' must be defined in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)
    isCopyOnNode = False
    if config.get(sectionName,"copy") == "yes" and config.get(sectionName,"tmpDir") != "":
        isCopyOnNode = True
            
    queueTmp = config.get(sectionName, "resources")
    cDir = os.getcwd()
    if config.get(sectionName, "tmpDir" ) != "":
        tmpDir = config.get(sectionName, "tmpDir")
    else:
        tmpDir = cDir
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    lCmdsTuples = []
    fileSize = float(os.path.getsize("%s/%s_db/%s_refTEs.fa" % (projectDir, projectName, projectName)) + 5000000) / 1000000000
    
    lCmdSize = []
    lCmdCopy = []
    if isCopyOnNode:
        lCmdSize.append("fileSize = %f" % fileSize) 
        lCmdCopy.append("shutil.copy(\"%s/%s_db/%s_refTEs.fa\", \".\")" % (projectDir, projectName, projectName))
        
    batchesDir = "%s/%s_db/batches%s" % (projectDir, projectName, qryDirSuffix)
    lFiles = FileUtils.getFileNamesList(batchesDir, "batch_")

    sectionName = "align_refTEs_with_genome"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
        
    if align == "BLR":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "BLR_blast") 
        except NoOptionError:
            print "ERROR: the option 'BLR_blast' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "BLR_sensitivity") 
        except NoOptionError:
            print "ERROR: the option 'BLR_sensitivity' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
        
        blastToUse = config.get(sectionName, "BLR_blast")
            
        if blastToUse not in ["ncbi", "blastplus", "wu"]:
            print "ERROR: 'BLR_blast: %s' in your configuration file is not authorized" % blastToUse
            sys.exit(1)
        
        if blastToUse == "blastplus":
            if not (CheckerUtils.isExecutableInUserPath("blastn") or CheckerUtils.isExecutableInUserPath("makeblastdb")):
                print "ERROR: ncbi blast plus (blastn executable) must be in your path"
                sys.exit(1)
            else:
                process = subprocess.Popen("blastn -h", shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
                output = process.communicate()[0]
                if "washu" in output.lower():
                    print "ERROR: you are trying to use NCBI BlastPlus, but the blastn executable is pointing to WuBlast. Please check your system path."
                    sys.exit(1)
        elif blastToUse == "wu":
            if not (CheckerUtils.isExecutableInUserPath("blastn") or CheckerUtils.isExecutableInUserPath("xdformat")):
                print "ERROR: WuBlast (blastn executable) must be in your path"
                sys.exit(1)
            else:
                process = subprocess.Popen("blastn -h", shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
                output = process.communicate()[0]
                if not "washu" in output.lower():
                    print "ERROR: you are trying to use WuBlast, but the blastn executable is pointing to NCBI BlastPlus. Please check your system path."
                    sys.exit(1)
        else:
            if not (CheckerUtils.isExecutableInUserPath("blastall") or CheckerUtils.isExecutableInUserPath("formatdb")):
                print "ERROR: ncbi blastall and formatdb must be in your path"
                sys.exit(1)
            
        groupid = "%s_TEannot_S2_%s%s" % (projectName, align, qryDirSuffix)
        acronym = "Blaster"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym, chooseTemplateWithCopy = isCopyOnNode)

        for file in lFiles:
            lCmds = []
            lCmds.append(getBlasterCmd(verbose, projectName, "align_refTEs_with_genome", config, iLauncher, file))
            lCmdStart = []
            if isCopyOnNode:
                lCmdStart.append("os.symlink(\"../%s_refTEs.fa\", \"%s_refTEs.fa\")" % (projectName, projectName))
                lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
            else:
                lCmdStart.append("os.symlink(\"%s/%s_db/%s_refTEs.fa\", \"%s_refTEs.fa\")" % (projectDir, projectName, projectName, projectName))
                lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (batchesDir, file, file))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.align\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.align\", \"%s/.\" )" % (file, cDir))
            lCmdFinish.append("shutil.move(\"%s.param\", \"%s/.\" )" % (file, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish, lCmdSize, lCmdCopy))
        
    elif align == "RM":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "RM_engine") 
        except NoOptionError:
            print "ERROR: the option 'RM_engine' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "RM_sensitivity") 
        except NoOptionError:
            print "ERROR: the option 'RM_sensitivity' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
            
        groupid = "%s_TEannot_S2_%s%s" % (projectName, align, qryDirSuffix)
        acronym = "RepeatMasker"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym, chooseTemplateWithCopy = isCopyOnNode)

        for file in lFiles:
            lCmdStart = []
            if isCopyOnNode:
                lCmdStart.append("os.symlink(\"../%s_refTEs.fa\", \"%s_refTEs.fa\")" % (projectName, projectName))
                lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
            else:
                lCmdStart.append("os.symlink(\"%s/%s_db/%s_refTEs.fa\", \"%s_refTEs.fa\")" % (projectDir, projectName, projectName, projectName))
                lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (batchesDir, file, file))
            lCmds = []
            lCmds.append(getRepeatMaskerCmd(projectName, "align_refTEs_with_genome", config, iLauncher, file))
            lCmds.append("if os.path.exists(\"%s.cat\"):" % file)
            lCmds.append("\t%s" % getRM2AlignCmd(iLauncher, "%s.cat" % file))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.cat\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.cat\", \"%s/.\" )" % (file, cDir))
            lCmdFinish.append("if os.path.exists(\"%s.cat.align\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.cat.align\", \"%s/.\" )" % (file, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish, lCmdSize, lCmdCopy))
        
    elif align == "CEN":

        groupid = "%s_TEannot_S2_%s%s" % (projectName, align, qryDirSuffix)
        acronym = "Censor"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym, chooseTemplateWithCopy = isCopyOnNode)

        for file in lFiles:
            lCmdStart = []
            if isCopyOnNode:
                lCmdStart.append("os.symlink(\"../%s_refTEs.fa\", \"%s_refTEs.fa\")" % (projectName, projectName))
                lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
            else:
                lCmdStart.append("os.symlink(\"%s/%s_db/%s_refTEs.fa\", \"%s_refTEs.fa\")" % (projectDir, projectName, projectName, projectName))
                lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (batchesDir, file, file))
            lCmds = []
            lCmds.append(getCensorCmd(projectName, sectionName, iLauncher, file))
            lCmds.append("if os.path.exists(\"%s.map\"):" % file)
            lCmds.append("\t%s" % getCensor2AlignCmd(iLauncher, "%s.map" % file))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.map\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.map\", \"%s/.\" )" % (file, cDir))
            lCmdFinish.append("if os.path.exists(\"%s.map.align\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.map.align\", \"%s/.\" )" % (file, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish, lCmdSize, lCmdCopy))
            
    else:
        print "ERROR: option -a %s not recognized" % ( align )
        sys.exit(1)
    
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean, isCopyOnNode)
    
    os.chdir( "../.." )
    
    if verbose > 0:
        print "step 2 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def getMatcherCmd(iLauncher, file):
    lArgs = []
    lArgs.append("-m %s" % file)
    lArgs.append("-j -x")
    return iLauncher.getSystemCommand("matcher", lArgs)
        
#------------------------------------------------------------------------------

def createFilterAlignCommand(alignFile, thres, iLauncher):
    lArgs = []
    lArgs.append("-i %s" % alignFile)
    lArgs.append("-S %s" % thres)
    lArgs.append("-r")
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("FilterAlign.py", lArgs)

def createFilterMatchWithUnknownNtCommand(iLauncher):
    lArgs = []
    lArgs.append("-p %s_chk_allTEs_tmpH_path" % projectName)
    lArgs.append("-s %s_chk_seq" % projectName)
    lArgs.append("-t 90")
    lArgs.append("-C %s" % configFileName)
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("FilterMatchWithUnknownNt.py", lArgs)

def createRetrieveInitHeadersCommand(iLauncher):
    lArgs = []
    lArgs.append("-i %s_chk_allTEs_tmpH_path" % projectName)
    lArgs.append("-l %s/%s_db/%s_refTEs.fa_initH.newHlink" % (projectDir, projectName, projectName))
    lArgs.append("-s")
    lArgs.append("-o %s_chk_allTEs_path" % projectName)
    lArgs.append("-c")
    return iLauncher.getSystemCommand("RetrieveInitHeaders.py", lArgs)

def createConvChunk2ChrCommand(iLauncher):
    lArgs = []
    lArgs.append("-m %s_chk_map" % projectName)
    lArgs.append("-q %s_chk_allTEs_path" % projectName)
    lArgs.append("-t path")
    lArgs.append("-c")
    lArgs.append("-o %s_chr_allTEs_path" % projectName)
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("srptConvChunk2Chr.py", lArgs)

def filterAndCombineTEsAnnot( combAlign ):
    """
    Step 3: filter the true HSPs with the randomized ones and combine them among the 3 programs
    """

    if verbose > 0:
        print "\nbeginning of step 3"
        print "filter the real HSPs with the random HSPs and combine them"
        sys.stdout.flush()

    if combAlign == "":
        print "ERROR: no alignment program specified with option -c"
        sys.exit(1)

    # retrieve for each program ran on the random chunks the highest score from all 'align' files
    thres = {}
    isExceptionRaised = False
    sectionName = "filter"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
                    
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "force_default_values") 
    except NoOptionError:
        print "ERROR: the option 'force_default_values' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
       
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "BLR") 
    except NoOptionError:
        print "ERROR: the option 'BLR' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
       
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "RM") 
    except NoOptionError:
        print "ERROR: the option 'RM' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
       
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "CEN") 
    except NoOptionError:
        print "ERROR: the option 'CEN' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True

    if isExceptionRaised: 
        sys.exit(1)

    if config.get( "filter", "force_default_values" ) == "yes":
        if verbose > 0:
            print "\n* force the use of the default values from the configuration file"
            sys.stdout.flush()
        for alignProgram in ["BLR","RM","CEN"]:
            thres[alignProgram] = config.get( "filter", alignProgram )
            if verbose > 0:
                print "default threshold %s: %s" % (alignProgram, thres[alignProgram])
                sys.stdout.flush()
    else:
        if os.path.exists( "%s_TEdetect_rnd" % projectName):
            os.chdir( "%s/%s_TEdetect_rnd/" % (projectDir, projectName) )
            cmd = "ComputeScoreThresholdFromAlignFiles.py -d %s -v %i" % (os.getcwd(), verbose - 1)
            process = subprocess.Popen(cmd, shell = True)
            process.communicate()
            if process.returncode != 0:
                raise Exception("ERROR when launching '%s'" % cmd)
            
            thresFile = "threshold.tmp"
            thresFileHandler = open(thresFile, "r")
            for line in thresFileHandler:
                lElt = line.split()
                thres[lElt[0]] = lElt[1]
            thresFileHandler.close()

    if thres == {}:
        print "ERROR: you have to launch step 2 with option \"-r\" or specify default filter values in the configuration file"
        sys.exit(1)

    os.chdir("%s/%s_TEdetect/" % (projectDir, projectName))
    
    lProg = combAlign.split("+")

    # filter the 'align' files
    if verbose > 0:
        print "\n* filter the 'align' files according to the threshold"; sys.stdout.flush()
        
    cDir = os.getcwd()
    if config.get(jobsSectionName, "tmpDir" ) != "":
        tmpDir = config.get(jobsSectionName, "tmpDir")
    else:
        tmpDir = cDir
    groupid = "%s_TEannot_step3_FilterAlign" % projectName
    acronym = "FilterAlign"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym)
    lCmdsTuples = []
    for prg in lProg:
        prgDir = "%s/%s" % (cDir, prg)
        lFiles = FileUtils.getFileNamesList(prgDir, ".*\.align")
        if not os.path.exists(prg):
            iDb.close()
            raise RepetException("ERROR: %s doesn't exist" % prgDir)
        elif not lFiles:
            iDb.close()
            raise RepetException("ERROR: no '.align' files in %s. Check the execution of step 2 with %s or don't run step 3 with %s." % (prgDir, prg, prg))
        else:
            lCmds = []
            lCmdStart = []
            lCmdFinish = []
            for alignFile in lFiles:
                cmd = createFilterAlignCommand(alignFile, thres[prg], iLauncher)
                lCmds.append(cmd)
                lCmdStart.append("os.symlink('%s/%s', '%s')" % (prgDir, alignFile, alignFile))
                lCmdFinish.append("shutil.move('%s.filtered', '%s/.')" % (alignFile, prgDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)
        
    # concatenate chunk by chunk the filtered 'align' files from each alignment program
    if verbose > 0:
        print "\n* concatenate chunk by chunk the filtered 'align' files from each alignment program"; sys.stdout.flush()
    if os.path.exists("Comb"):
        shutil.rmtree("Comb")
    os.mkdir("Comb")
    os.chdir("Comb")
    lChkFilePaths = glob.glob( "%s/%s_db/batches/batch*.fa" % (projectDir, projectName))
    for chkFilePath in lChkFilePaths:
        chkFileName = os.path.basename(chkFilePath)
        concatFile = "%s_allTEs.align" % chkFileName
        prg = "cat"
        cmd = prg
        if "BLR" in combAlign and os.path.exists( "../BLR/%s.align.filtered" % chkFileName ):
            if "RM" in combAlign and os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered ../RM/%s.cat.align.filtered ../CEN/%s.map.align.filtered > %s" % ( chkFileName, chkFileName, chkFileName, concatFile )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered ../RM/%s.cat.align.filtered > %s" % ( chkFileName, chkFileName, concatFile )
            elif "RM" not in combAlign or not os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered ../CEN/%s.map.align.filtered > %s" % ( chkFileName, chkFileName, concatFile )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../BLR/%s.align.filtered > %s" % ( chkFileName, concatFile )
        elif "BLR" not in combAlign or not os.path.exists( "../BLR/%s.align.filtered" % chkFileName ):
            if "RM" in combAlign and os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../RM/%s.cat.align.filtered ../CEN/%s.map.align.filtered > %s" % ( chkFileName, chkFileName, concatFile )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../RM/%s.cat.align.filtered > %s" % ( chkFileName, concatFile )
            elif "RM" not in combAlign or not os.path.exists( "../RM/%s.cat.align.filtered" % chkFileName ):
                if "CEN" in combAlign and os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    cmd += " ../CEN/%s.map.align.filtered > %s" % ( chkFileName, concatFile )
                elif "CEN" not in combAlign or not os.path.exists( "../CEN/%s.map.align.filtered" % chkFileName ):
                    continue
#        for prg in lProg:
#            if os.path.exists("../%s/%s.*align.filtered" % (prg, chkFileName)):
#                cmd += " ../%s/%s.*align.filtered" % (prg, chkFileName)
#        if cmd != prg:
#            cmd += " > %s" % concatFile
#            process = subprocess.Popen(cmd, shell = True)
#            process.communicate()
#            if process.returncode != 0:
#                raise Exception("ERROR when launching '%s'" % cmd)
#        
#            if len(lProg) > 1:
#                prg = "UpdateScoreOfMatches.py"
#                cmd = prg
#                cmd += " -i %s" % concatFile
#                cmd += " -f align"
#                cmd += " -o %s.newScores" % concatFile
#                cmd += " -v 0"
#                process = subprocess.Popen(cmd, shell = True)
#                process.communicate()
#                if process.returncode != 0:
#                    raise Exception("ERROR when launching '%s'" % cmd)
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)
        
        if len(combAlign) > 5:
            prg = "UpdateScoreOfMatches.py"
            cmd = prg
            cmd += " -i %s" % concatFile
            cmd += " -f align"
            cmd += " -o %s.newScores" % concatFile
            cmd += " -v 0"
            process = subprocess.Popen(cmd, shell = True)
            process.communicate()
            if process.returncode != 0:
                raise Exception("ERROR when launching '%s'" % cmd)
            
    # join the matches with Matcher
    if verbose > 0:
        print "\n* join the matches with Matcher"
        sys.stdout.flush()
            
    cDir = os.getcwd()
    if config.get(jobsSectionName, "tmpDir" ) != "":
        tmpDir = config.get(jobsSectionName, "tmpDir")
    else:
        tmpDir = cDir
        
    groupid = "%s_TEannot_S3_Matcher" % projectName
    acronym = "Matcher"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym)

    lCmdsTuples = []
    lFiles = FileUtils.getFileNamesList(cDir, ".*\.newScores")
    for file in lFiles:
        lCmdStart = []
        lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (cDir, file))
        lCmds = []
        lCmds.append(getMatcherCmd(iLauncher, file))
        lCmdFinish = []
        lCmdFinish.append("if os.path.exists(\"%s.clean_match.map\"):" % file)
        lCmdFinish.append("\tshutil.move(\"%s.clean_match.map\", \"%s/.\" )" % (file, cDir))
        lCmdFinish.append("if os.path.exists(\"%s.clean_match.path\"):" % file)
        lCmdFinish.append("\tshutil.move(\"%s.clean_match.path\", \"%s/.\" )" % (file, cDir))
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
    
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)
    iMMF = MergeMatchsFiles("path", "%s_TEannot_Matcher" % projectName, clean = clean)
    iMMF.run()
            
    # filter and format data
    if verbose > 0:
        print "\n* filter and format data"
        sys.stdout.flush()
    
    iDb = DbFactory.createInstance()
    iDb.createTable("%s_chk_allTEs_tmpH_path" % projectName, "path", "%s_TEannot_Matcher.path" % projectName, True)
    iDb.close()
        
    groupid = "%s_TEannot_step3_FilterUnknownNt_RetrieveInitHeaders_ConvChkToChr" % projectName
    acronym = "FilterMatchWithUnknownNt_RetrieveInitHeaders_srptConvChunk2Chr"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym)
    
    lCmds = []
    # Remove matches containing more than 90% of 'N' nucleotides
    lCmds.append(createFilterMatchWithUnknownNtCommand(iLauncher))
    # retrieve the initial headers of the reference TEs
    lCmds.append(createRetrieveInitHeadersCommand(iLauncher))
    # convert the coordinates from chunks to chromosomes
    lCmds.append(createConvChunk2ChrCommand(iLauncher))

    lCmdsTuples = []
    lCmdStart = []
    lCmdStart.append("os.symlink('%s/%s', '%s')" % (cDir, configFileName, configFileName))
    lCmdStart.append("os.symlink('%s/%s_db/%s_refTEs.fa_initH.newHlink', '%s_refTEs.fa_initH.newHlink')" % (projectDir, projectName, projectName, projectName))
    lCmdStart.append("os.symlink('%s/%s_TEannot_Matcher.path', '%s_TEannot_Matcher.path')" % (cDir, projectName, projectName))
    lCmdFinish = []
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)
    
    if clean:
        cmd = "find . -name \"*.align\" -exec rm {} \;"
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing temporary 'align' files"
            sys.exit(1)
        cmd = "find . -name \"*.align.clean_match.map\" -exec rm {} \;"
        log = os.system( cmd )
        if log != 0:
            print "ERROR while removing temporary 'map' files"
            sys.exit(1)

    os.chdir( "../.." )

    if verbose > 0:
        print "step 3 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def getLaunchTRFCmd(sectionName,config, iLauncher, file):
    lArgs = []
    lArgs.append("-i %s" % file)
    lArgs.append("-c")
    lArgs.append("-m %s" % config.get(sectionName, "TRFmaxPeriod"))
        
    return iLauncher.getSystemCommand("LaunchTRF.py", lArgs)

def getRMSSRCmd(sectionName, config, iLauncher, file):
    lArgs = []
#    lArgs.append("-dir . -pa 1 -gccalc -no_is -int -s -e wublast")
    lArgs.append("-dir . -pa 1 -gccalc -no_is -int -s")
    if config.get(sectionName, "RMSSR_engine") == "wu": 
        lArgs.append("-e wublast")
    elif config.get(sectionName, "RMSSR_engine") == "cm":
        lArgs.append("-e crossmatch")
    elif config.get(sectionName, "RMSSR_engine") == "ncbi":
        lArgs.append("-e ncbi")

    lArgs.append(file)
    return iLauncher.getSystemCommand("RepeatMasker", lArgs)

def getRM2PathCmd(iLauncher, file):
    lArgs = []
    lArgs.append("-i %s.cat" % file)
    return iLauncher.getSystemCommand("RMcat2path.py", lArgs)

def getLaunchMrepsCmd(iLauncher, file):
    lArgs = []
    lArgs.append("-i %s" % file)
    lArgs.append("-c")
    return iLauncher.getSystemCommand("launchMreps.py", lArgs)

#------------------------------------------------------------------------------

def detectSSRs( sat ):

    """
    Step 4: detect SSRs

    @param sat: SSR detection program (TRF, RMSSR, Mreps)
    @type sat: string
    """

    if verbose > 0:
        print "\nbeginning of step 4"
        print "search for satellites on the chunks with %s" %( sat )
        sys.stdout.flush()

    if not os.path.exists( "%s_SSRdetect" % ( projectName ) ):
        os.mkdir( "%s_SSRdetect" % ( projectName ) )
    os.chdir( "%s_SSRdetect" % ( projectName ) )
    if not os.path.exists( sat ):
        os.mkdir( sat )
    else:
        print "ERROR: directory %s_SSRdetect/%s/ already exists" % ( projectName, sat )
        sys.exit(1)
    os.chdir( sat )

    sectionName = "SSR_detect"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)

    # launch the SSR-detection program (TRF or Mreps or RM) on  all the chunks in parallel
    if sat ==  "TRF" and not CheckerUtils.isExecutableInUserPath("trf"):
        print "ERROR: trf must be in your path"
        sys.exit(1)
    if sat ==  "RMSSR" and not CheckerUtils.isExecutableInUserPath("RepeatMasker"):
        print "ERROR: RepeatMasker must be in your path"
        sys.exit(1)
    if sat ==  "Mreps" and not CheckerUtils.isExecutableInUserPath("mreps"):
        print "ERROR: mreps must be in your path"
        sys.exit(1)
            
    sectionName = "parallelized_jobs"
    queueTmp = config.get(sectionName, "resources")
    cDir = os.getcwd()
    if config.get(sectionName, "tmpDir" ) != "":
        tmpDir = config.get(sectionName, "tmpDir")
    else:
        tmpDir = cDir
        
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    batchesDir = "%s/%s_db/batches" % (projectDir, projectName)
    lFiles = FileUtils.getFileNamesList(batchesDir, "batch_")
    lCmdsTuples = []

    if sat == "TRF":
        type = "set"
        groupid = "%s_TEannot_S4_TRF" % projectName
        acronym = "TRF"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym)
        for file in lFiles:
            lCmdStart = []
            lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
            lCmds = []
            lCmds.append(getLaunchTRFCmd("SSR_detect", config, iLauncher, file))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.TRF.set\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.TRF.set\", \"%s/.\" )" % (file, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        
    elif sat == "RMSSR":
        type = "path"
        groupid = "%s_TEannot_S4_RMSSR" % projectName
        acronym = "RMSSR"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym)
        for file in lFiles:
            lCmdStart = []
            lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
            lCmds = []
            lCmds.append(getRMSSRCmd("SSR_detect", config, iLauncher, file))
            lCmds.append("if os.path.exists(\"%s.cat\"):" % file)
            lCmds.append("\t%s" % getRM2PathCmd(iLauncher, file))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.cat.path\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.cat.path\", \"%s/%s.RMSSR.path\" )" % (file, cDir, file))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        
    elif sat == "Mreps":
        type = "set"
        groupid = "%s_TEannot_S4_Mreps" % projectName
        acronym = "Mreps"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym)
        for file in lFiles:
            lCmdStart = []
            lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
            lCmds = []
            lCmds.append(getLaunchMrepsCmd(iLauncher, file))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.Mreps.set\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.Mreps.set\", \"%s/.\" )" % (file, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))

    else:
        print "ERROR: option -s %s not recognized" % ( sat )
        sys.exit(1)
        
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)
    iMMF = MergeMatchsFiles(type, "%s_TEannot_%s" % (projectName, sat), clean = clean)
    iMMF.run()

    # load the results into a MySQL table
    tableName = "%s_chk_%s_" % (projectName, sat)
    fileName = "%s_TEannot_%s." % (projectName, sat)
    iDb = DbFactory.createInstance()
    iDb.createTable("%s%s" % (tableName, type), type, "%s%s" % (fileName, type), True)
    iDb.close()

    os.chdir( ".." )

    if verbose > 0:
        print "step 4 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def filterAndCombineSSRsAnnot():

    """
    Step 5: merge the SSR annotations from TRF, Mreps and RepeatMasker
    """

    if verbose > 0:
        print "\nbeginning of step 5"
        print "filter and combine the SSR annotations from TRF, Mreps and RepeatMasker"
        sys.stdout.flush()

    if not os.path.exists( "%s_SSRdetect" % ( projectName ) ):
        print "ERROR: directory %s_SSRdetect doesn't exist" % ( projectName )
        sys.exit(1)
    os.chdir( "%s_SSRdetect" % ( projectName ) )
    if os.path.exists( "Comb" ):
        print "ERROR: directory %s_SSRdetect/Comb already exists" % ( projectName )
        sys.exit(1)
    os.mkdir( "Comb" )
    os.chdir( "Comb" )

    #TODO: choose analysis with option instead of renaming a directory to avoid an analysis
    mergeFile = open( "mergeSSR.txt", "w" )
    cmd = ""
    if os.path.exists( "../TRF" ):
        cmd += "%s_chk_TRF_set\tset\tmerge\n" % ( projectName )
    if os.path.exists( "../Mreps" ):
        cmd += "%s_chk_Mreps_set\tset\tmerge\n" % ( projectName )
    if os.path.exists( "../RMSSR" ):
        cmd += "%s_chk_RMSSR_path\tpath\tmerge\n" % ( projectName )
    mergeFile.write( cmd )
    mergeFile.close()
        
    cDir = os.getcwd()
    if config.get(jobsSectionName, "tmpDir" ) != "":
        tmpDir = config.get(jobsSectionName, "tmpDir")
    else:
        tmpDir = cDir
    groupid = "%s_TEannot_S5_MergeSSRannotationAndConvChunk2Chr" % projectName
    acronym = "srptAutoPromote_srptConvChunk2Chr"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid)

    lCmds = []
    lArgs = []
    lArgs.append("-f mergeSSR.txt")
    lArgs.append("-o %s_chk_allSSRs_set" % projectName)
    lArgs.append("-v %i" % (verbose - 1))
    cmd = iLauncher.getSystemCommand("srptAutoPromote.py", lArgs)
    lCmds.append(cmd)

    lArgs = []
    lArgs.append("-m %s_chk_map" % projectName)
    lArgs.append("-q %s_chk_allSSRs_set" % projectName)
    lArgs.append("-t set")
    lArgs.append("-c")
    lArgs.append("-o %s_chr_allSSRs_set" % projectName)
    lArgs.append("-v %i" % (verbose - 1))
    cmd = iLauncher.getSystemCommand("srptConvChunk2Chr.py", lArgs)
    lCmds.append(cmd)
    
    lCmdsTuples = []
    lCmdStart = []
    lCmdStart.append("os.symlink(\"%s/mergeSSR.txt\", \"mergeSSR.txt\")" % cDir)
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)

    os.chdir( "../.." )

    if verbose > 0:
        print "step 5 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def getBlasterCmdWithBank(bankName, typeBlast, iLauncher, file):
    lArgs = []
    lArgs.append("-q %s" % file)
    lArgs.append("-s %s" % bankName)
    lArgs.append("-B %s" % file)
    lArgs.append("-n %s -r" % typeBlast)
    #TODO: when Wublast is launched, don't use more than 1 CPU
#    if config.get(sectionName, "BLR_blast") == "ncbi": 
#        lArgs.append("-N")
#        lArgs.append("-p '-a 1'")
#    elif config.get(sectionName, "BLR_blast") == "wu":
#        lArgs.append("-W")
#        lArgs.append("-p -cpus=1")
#    elif config.get(sectionName, "BLR_blast") == "blastplus":
#        lArgs.append("-X")
#        lArgs.append("-p '-num_threads 1'")
    return iLauncher.getSystemCommand("blaster", lArgs)

def getMatcherCmdWithBank(bankName, iLauncher, file):
    lArgs = []
    lArgs.append("-m %s.align" % file)
    lArgs.append("-q %s" % file)
    lArgs.append("-s %s" % bankName)
    lArgs.append("-j")
    return iLauncher.getSystemCommand("matcher", lArgs)

#------------------------------------------------------------------------------

def alignChunksAndTEsForCoding( typeBlast ):

    """
    Step 6: align the 2nd data bank on the genomic sequences via Blaster with tblastx or blastx
    * load the bank in a MySQL table
    * prepare the bank
    * launch BLASTER with tblastx or blastx
    """

    if verbose > 0:
        print "\nbeginning of step 6"; sys.stdout.flush()

    if not os.path.exists( "%s_db" % ( projectName ) ):
        os.mkdir( "%s_db" % ( projectName ) )
    os.chdir( "%s_db" % ( projectName ) )
    typeSeq = ""
    
    if typeBlast == "tblastx":
        bankBlast = "bankBLRtx"
        typeSeq = "nt"
    elif typeBlast == "blastx":
        bankBlast = "bankBLRx"
        typeSeq = "prot"

    sectionName = "align_other_banks"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, bankBlast) 
    except NoOptionError:
        print "ERROR: the option '%s' must be defined in '%s' in your config file : %s" % (bankBlast, sectionName, configFileName)
        sys.exit(1)

    bankName = config.get(sectionName, bankBlast)
    if not os.path.exists( "%s/%s" % (projectDir, bankName) ):
        print "ERROR: %s/%s doesn't exist" % (projectDir, bankName)
        sys.exit(1)

    #TODO: usefull ?
    if not os.path.exists( "%s/%s_db/%s" % (projectDir, projectName, bankName) ):
        os.system( "ln -s ../%s %s" % (bankName, bankName) )
        
    FileUtils.fromWindowsToUnixEof( bankName )
    
    # load the bank in a MySQL table
    iDb = DbFactory.createInstance()
    iDb.createTable("%s_%s_%s_seq" % ( projectName, bankBlast, typeSeq ), "fasta", bankName, True)
    iDb.close()

    # prepare the bank
    cmd = "blaster -q %s -n %s -B %s-%sPrepare -P" % (bankName, typeBlast, bankName, bankBlast)
    process = subprocess.Popen(cmd, shell = True)
    process.communicate()
    if process.returncode != 0:
        raise Exception("ERROR when launching '%s'" % cmd)

    os.chdir("..")

    #--------------------------------------------------------------------------

    if verbose > 0:
        print "\nalign '%s' and the chunks via Blaster with %s" % ( bankName, typeBlast )
        sys.stdout.flush()

    if not os.path.exists( projectName + "_TEdetect" ):
        os.mkdir( projectName + "_TEdetect" )
    os.chdir( projectName + "_TEdetect" )
    if not os.path.exists( bankBlast ):
        os.mkdir( bankBlast )
    else:
        print "ERROR: directory %s_TEdetect/%s/ already exists" % ( projectName, bankBlast )
        sys.exit(1)
    os.chdir( bankBlast )

    # launch Blaster with tblastx followed by Matcher on each chunk in parallel
    sectionName = "parallelized_jobs"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "copy")
    except NoOptionError:
        print "ERROR: the option 'copy' must be defined in " + sectionName +" in your config file : " + configFileName
        sys.exit(1)
    isCopyOnNode = False
    if config.get(sectionName,"copy") == "yes" and config.get(sectionName,"tmpDir") != "":
        isCopyOnNode = True
        
    queueTmp = config.get(sectionName, "resources")
    cDir = os.getcwd()
    if config.get(sectionName, "tmpDir" ) != "":
        tmpDir = config.get(sectionName, "tmpDir")
    else:
        tmpDir = cDir
    fileSize = float(os.path.getsize("%s/%s_db/%s" % (projectDir, projectName, bankName)) + 5000000) / 1000000000
    
    lCmdSize = []
    lCmdCopy = []
    if isCopyOnNode:
        lCmdSize.append("fileSize = %f" % fileSize) 
        lCmdCopy.append("shutil.copy(\"%s/%s_db/%s\", \".\")" % (projectDir, projectName, bankName))
        
    groupid = "%s_TEannot_S6_%s" % (projectName, typeBlast)
    acronym = "BlasterMatcher"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queueTmp, groupid, acronym, chooseTemplateWithCopy = isCopyOnNode)
    
    batchesDir = "%s/%s_db/batches" % (projectDir, projectName)
    lFiles = FileUtils.getFileNamesList(batchesDir, "batch_")
    lCmdsTuples = []
    for file in lFiles:
        lCmds = []
        lCmds.append(getBlasterCmdWithBank(bankName, typeBlast, iLauncher, file))
        lCmds.append(getMatcherCmdWithBank(bankName, iLauncher, file))
        lCmdStart = []
        if isCopyOnNode:
            lCmdStart.append("os.symlink(\"../%s\", \"%s\")" % (bankName, bankName))
            lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (batchesDir, file))
        else:
            lCmdStart.append("os.symlink(\"%s/%s_db/%s\", \"%s\")" % (projectDir, projectName, bankName, bankName))
            lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (batchesDir, file, file))
        lCmdFinish = []
        lCmdFinish.append("if os.path.exists(\"%s.param\"):" % file)
        lCmdFinish.append("\tshutil.move(\"%s.param\", \"%s/.\" )" % (file, cDir))
        lCmdFinish.append("if os.path.exists(\"%s.align.clean_match.path\"):" % file)
        lCmdFinish.append("\tshutil.move(\"%s.align.clean_match.path\", \"%s/.\")" % (file, cDir))
        lCmdFinish.append("if os.path.exists(\"%s.align.clean_match.param\"):" % file)
        lCmdFinish.append("\tshutil.move(\"%s.align.clean_match.param\", \"%s/.\" )" % (file, cDir))
        if not clean:
            lCmdFinish.append("if os.path.exists(\"%s.align\"):" % file)
            lCmdFinish.append("\tshutil.move(\"%s.align\", \"%s/.\")" % (file, cDir))
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish, lCmdSize, lCmdCopy))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean, isCopyOnNode)
    iMMF = MergeMatchsFiles("path", "%s_%s" % (projectName, bankBlast), clean = clean)
    iMMF.run()

    # load data in MySQL
    iDb = DbFactory.createInstance()
    iDb.createTable("%s_chk_%s_path" % (projectName, bankBlast), "path", "%s_%s.path" % (projectName, bankBlast), True)
    iDb.close()
    
    # convert the coordinates from chunks to chromosomes
    groupid = "%s_TEannot_S6_convChunk2Chr" % projectName
    acronym = "srptConvChunk2Chr_%s" % typeBlast
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid)
    
    lArgs = []
    lArgs.append("-m %s_chk_map" % projectName)
    lArgs.append("-q %s_chk_%s_path" % (projectName, bankBlast))
    lArgs.append("-t path")
    lArgs.append("-c")
    lArgs.append("-o %s_chr_%s_path" % (projectName, bankBlast))
    lArgs.append("-v %i" % (verbose - 1))
    cmd = iLauncher.getSystemCommand("srptConvChunk2Chr.py", lArgs)
    
    lCmdsTuples = []
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation([cmd]))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)    

    os.chdir( ".." )

    if verbose > 0:
        print "step 6 finished successfully\n"; sys.stdout.flush()

#------------------------------------------------------------------------------

def pathsProcessing( config ):

    """
    Step 7: remove path doublons, spurious paths and apply long join procedure
    """

    if verbose > 0:
        print "\nbeginning of step 7"
        sys.stdout.flush()

    os.chdir( "%s_TEdetect" % ( projectName ) )

    isExceptionRaised = False
    sectionName = "annot_processing"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_size") 
    except NoOptionError:
        print "ERROR: the option 'min_size' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_id_tolerance") 
    except NoOptionError:
        print "ERROR: the option 'join_id_tolerance' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_max_gap_size") 
    except NoOptionError:
        print "ERROR: the option 'join_max_gap_size' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_max_mismatch_size") 
    except NoOptionError:
        print "ERROR: the option 'join_max_mismatch_size' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_TEinsert_cov") 
    except NoOptionError:
        print "ERROR: the option 'join_TEinsert_cov' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_overlap") 
    except NoOptionError:
        print "ERROR: the option 'join_overlap' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "join_minlength_split") 
    except NoOptionError:
        print "ERROR: the option 'join_minlength_split' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    if isExceptionRaised == True:
        sys.exit(1)
    
    cDir = os.getcwd()
    if config.get(jobsSectionName, "tmpDir" ) != "":
        tmpDir = config.get(jobsSectionName, "tmpDir")
    else:
        tmpDir = cDir

    # remove doublons and spurious hits
    #TODO: how to have comments at the correct moment ? 
    if verbose > 0:
        print "remove path doublons among the TE annotations"
        print "remove TE annotations corresponding to SSRs"
        sys.stdout.flush()
    groupid = "%s_TEannot_S7_RemoveDuplicatesAndSpuriousPaths" % projectName
    acronym = "srptRemovePathDoublons_srptRemoveSpurious"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid)

    lCmds = []
    lArgs = []
    lArgs.append("-i %s_chk_allTEs_path" % projectName)
    lArgs.append("-o %s_chk_allTEs_nr_path" % projectName)
    lArgs.append("-v %i" % (verbose - 1))
    cmd = iLauncher.getSystemCommand("srptRemovePathDoublons.py", lArgs)
    lCmds.append(cmd)

    iDb = DbFactory.createInstance("%s/%s" % (projectDir, configFileName)) 
    if iDb.doesTableExist("%s_chk_allSSRs_set" % projectName):
        lArgs = []
        lArgs.append("-q %s_chk_allTEs_nr_path" % projectName)
        lArgs.append("-s %s_chk_allSSRs_set" % projectName)
        lArgs.append("-t path/set")
        lArgs.append("-r %s" % config.get(sectionName, "min_size"))
        lArgs.append("-o %s_chk_allTEs_nr_noSSR_path" % projectName)
        lArgs.append("-v %i" % (verbose - 1))
        cmd = iLauncher.getSystemCommand("srptRemoveSpurious.py", lArgs)
        lCmds.append(cmd)
    else:
        print "WARNING: can't find table '%s_chk_allSSRs_set" % projectName
        sys.stdout.flush()
        cmd = "from commons.core.sql.DbFactory import DbFactory\n"
        cmd += "iDb = DbFactory.createInstance()\n"
        cmd += "iDb.copyTable( \"%s_chk_allTEs_nr_path\", \"%s_chk_allTEs_nr_noSSR_path\" )\n" % (projectName, projectName)
        cmd += "iDb.close()\n"
        lCmds.append(cmd)
    
    # avoid time out on iDb, close db handler
    iDb.close()
    
    lCmdsTuples = []
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)    

    # convert the coordinates from chunks to chromosomes
    groupid = "%s_TEannot_S7_ConvChunk2Chr" % projectName
    acronym = "srptConvChunk2Chr"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid)

    lArgs = []
    lArgs.append("-m %s_chk_map" % projectName)
    lArgs.append("-q %s_chk_allTEs_nr_noSSR_path" % projectName)
    lArgs.append("-t path")
    lArgs.append("-c")
    lArgs.append("-o %s_chr_allTEs_nr_noSSR_path" % projectName)
    lArgs.append("-v %i" % (verbose - 1))
    cmd = iLauncher.getSystemCommand("srptConvChunk2Chr.py", lArgs)
    
    lCmdsTuples = []
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation([cmd]))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)    
    
    # long join procedure to connect fragmented, nested TE copies
    if verbose > 0:
        print "long join procedure to connect fragmented TE copies"
        sys.stdout.flush()
    groupid = "%s_TEannot_S7_LongJoinsForTEs" % projectName
    acronym = "LongJoinsForTEs"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid)
    
    lArgs = []
    lArgs.append("-t %s_chr_allTEs_nr_noSSR_path" % projectName)
    lArgs.append("-i %s" % config.get(sectionName, "join_id_tolerance"))
    lArgs.append("-g %s" % config.get(sectionName, "join_max_gap_size"))
    lArgs.append("-m %s" % config.get(sectionName, "join_max_mismatch_size"))
    lArgs.append("-c %s" % config.get(sectionName, "join_TEinsert_cov"))
    lArgs.append("-o %s" % config.get(sectionName, "join_overlap"))
    lArgs.append("-s %s" % config.get(sectionName, "join_minlength_split"))
    lArgs.append("-l %s" % config.get(sectionName, "min_size"))
    lArgs.append("-O %s_chr_allTEs_nr_noSSR_join_path" % projectName)
    lArgs.append("-C %s/%s" % (projectDir, configFileName))
    lArgs.append("-v %i" % verbose)
    cmd = iLauncher.getSystemCommand("LongJoinsForTEs.py", lArgs)
    
    lCmdsTuples = []
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation([cmd]))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)
    
    # end of long join procedure, re open db connection
    if clean:
        iDb = DbFactory.createInstance("%s/%s" % (projectDir, configFileName))
        iDb.dropTable("%s_chr_allTEs_nr_noSSR_path" % projectName)
        iDb.close()
    
    # convert the coordinates from chromosomes to chunks
    groupid = "%s_TEannot_S7_ConvChr2Chunk" % ( projectName )
    acronym = "srptConvChr2Chunk"
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid)
    
    lArgs = []
    lArgs.append("-m %s_chk_map" % projectName)
    lArgs.append("-q %s_chr_allTEs_nr_noSSR_join_path" % projectName)
    lArgs.append("-t path")
    lArgs.append("-o %s_chk_allTEs_nr_noSSR_join_path" % projectName)
    cmd = iLauncher.getSystemCommand("srptConvChr2Chunk.py", lArgs)
    
    lCmdsTuples = []
    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation([cmd]))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)    

    os.chdir( ".." )

    if verbose > 0:
        print "step 7 finished successfully\n"
        sys.stdout.flush()

#------------------------------------------------------------------------------

def _createGFF3MakerCommand(sectionName, tablesFileName, classifTableName, iLauncher):
    lArgs = []
    lArgs.append("-t %s" % tablesFileName)
    lArgs.append("-f %s_chr_seq" % projectName)
    if config.get(sectionName, "gff3_chado") == "yes":
        lArgs.append("-c")
    if config.get(sectionName, "gff3_with_genomic_sequence") == "yes":
        lArgs.append("-w")
    if config.get(sectionName, "keep_gff3_files_without_annotations") == "yes":
        lArgs.append("-a")
    if config.get(sectionName, "gff3_merge_redundant_features") == "yes":
        lArgs.append("-m")
    if config.get(sectionName, "gff3_compulsory_match_part") == "yes":
        lArgs.append("-p")
    if classifTableName != "":
        lArgs.append("-i %s" % classifTableName)
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("GFF3Maker.py", lArgs)

def exportAnnotations( outFormat ):
    """
    Step 8: export files recording the annotations for each chunk

    @outFormat: gameXML or GFF3
    @type outFormat: string
    """

    if verbose > 0:
        print "\nbeginning of step 8"
        print "build %s files" % ( outFormat )
        sys.stdout.flush()

    isExceptionRaised = False
    sectionName = "export"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "sequences") 
    except NoOptionError:
        print "ERROR: the option 'sequences' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "add_SSRs") 
    except NoOptionError:
        print "ERROR: the option 'add_SSRs' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
        
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "add_tBx") 
    except NoOptionError:
        print "ERROR: the option 'add_tBx' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        isExceptionRaised = True
       
    if outFormat.lower() not in ["gff3", "gamexml"]: 
        print "ERROR: output format '%s' ('-o' option) not recognized. Possible values are 'GFF3' and 'gameXML'." % outFormat
        isExceptionRaised = True
        
    if isExceptionRaised == True:
        sys.exit(1)

    refSeq = config.get(sectionName, "sequences")
    if refSeq == "chromosomes":
        refSeq = "chr"
    elif refSeq == "chunks":
        refSeq = "chk"
    else:
        print "ERROR: '%s' not recognized"
        sys.exit(1)

    if not os.path.exists( "%s_%s%s" % (projectName, outFormat, refSeq) ):
        os.mkdir( "%s_%s%s" % (projectName, outFormat, refSeq) )
    os.chdir( "%s_%s%s" % (projectName, outFormat, refSeq) )

    addTElength = config.get(sectionName, "gff3_with_TE_length") == "yes"

    # write the file with the annotation tables
    tablesFileName = "annotation_tables.txt"
    tableFile = open( tablesFileName, "w" )
    string = "%s_REPET_TEs\t" % projectName
    string += "path\t"
    string += "%s_%s_allTEs_nr_noSSR_join_path" % (projectName, refSeq)
    if addTElength:
        string += "\t%s_refTEs_seq" % projectName
    string += "\n"
    if config.get(sectionName, "add_SSRs") == "yes":
        if os.path.exists( "%s/%s_SSRdetect/Comb" % (projectDir, projectName) ): 
            string += "%s_REPET_SSRs\t" % projectName
            string += "set\t"
            string += "%s_%s_allSSRs_set\n" % (projectName, refSeq)
        else:
            print "WARNING: can't export SSR annotations, launch step 4 and 5 first"
    if config.get(sectionName, "add_tBx") == "yes":
        if os.path.exists( "%s/%s_TEdetect/bankBLRtx" % (projectDir, projectName) ):
            string += "%s_REPET_tblastx\t" % projectName
            string += "path\t"
            string += "%s_%s_bankBLRtx_path" % (projectName, refSeq)
            if addTElength:
                string += "\t%s_bankBLRtx_nt_seq" % projectName
            string += "\n"
        else:
            print "WARNING: can't export tblastx annotations, launch step 6 first"
    if config.get(sectionName, "add_Bx") == "yes":
        if os.path.exists( "%s/%s_TEdetect/bankBLRx" % (projectDir, projectName) ):
            string += "%s_REPET_blastx\t" % projectName
            string += "path\t"
            string += "%s_%s_bankBLRx_path" % (projectName, refSeq)
            if addTElength:
                string += "\t%s_bankBLRx_prot_seq" % projectName
            string += "\n"
        else:
            print "WARNING: can't export blastx annotations, launch step 6 first"
    tableFile.write(string)
    tableFile.close()

    #--------------------------------------------------------------------------
    if outFormat.lower() == "gff3":
        
        classifTableName = ""
        iDb = DbFactory.createInstance()
        
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "gff3_chado") 
        except NoOptionError:
            print "ERROR: the option gff3_chado must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
            
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "gff3_with_classif_info") 
        except NoOptionError:
            print "ERROR: the option gff3_with_classif_info must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
            
        if config.get(sectionName, "gff3_with_classif_info") == "yes":
            if config.get(sectionName, "classif_table_name") != "":
                classifTableName = config.get(sectionName, "classif_table_name")
            else:
                classifTableName = "%s_consensus_classif" % projectName
            
            if not iDb.doesTableExist(classifTableName):
                print "ERROR: classification table '%s' doesn't exist!" % classifTableName
                sys.exit(1)

        # Write the GFF3 files
        if verbose > 0:
            print "Write GFF3 files..."
            sys.stdout.flush()
    
        cDir = os.getcwd()
        if config.get(jobsSectionName, "tmpDir" ) != "":
            tmpDir = config.get(jobsSectionName, "tmpDir")
        else:
            tmpDir = cDir
            
        
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        groupid = "%s_TEannot_step8_GFF3" % ( projectName )
        acronym = "GFF3Maker"
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym)
        lCmds = []
        lCmds.append(_createGFF3MakerCommand(sectionName, tablesFileName, classifTableName, iLauncher))
        lCmdStart = []
        lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (cDir, tablesFileName))
        lCmdFinish = []
        lCmdFinish.append("os.system(\"find . -type f -name '*.gff3' -exec mv {} %s/. \\;\")" % cDir)
        lCmdsTuples = []
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, clean)
            
        os.chdir( ".." )

    #--------------------------------------------------------------------------

    elif outFormat.lower() == "gamexml":

        # create the source gameXML file(s), one per genomic sequence
        if verbose > 0:
            print "create source files..."
            sys.stdout.flush()

        prg = "GameXmlMaker.py"
        cmd = prg
        if refSeq == "chr":
            cmd += " -f %s.fa" % projectName
        elif refSeq == "chk":
            cmd += " -f %s_chunks.fa" % projectName
        cmd += " -v %i" % ( verbose - 1 )
        
        cDir = os.getcwd()
        if config.get(jobsSectionName, "tmpDir" ) != "":
            tmpDir = config.get(jobsSectionName, "tmpDir")
        else:
            tmpDir = cDir

        groupid = "%s_TEannot_step8_GameXml_createFiles" % ( projectName )
        acronym = "GameXmlMaker"
        jobdb = RepetJob( cfgFileName = "%s/%s" % (projectDir, configFileName) )
        cL = Launcher( jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym )
        cL.beginRun()
        cL.job.jobname = acronym
        cmd_start = ""
        if refSeq == "chr":
            cmd_start += "os.system( \"ln -s %s/%s.fa %s.fa\" )\n" % (projectDir, projectName, projectName)
        elif refSeq == "chk":
            cmd_start += "os.system( \"ln -s %s/%s_db/%s_chunks.fa %s_chunks.fa\" )\n" % (projectDir, projectName, projectName, projectName)
        cmd_start += "log = os.system( \""
        cmd_start += cmd
        cmd_start += "\" )\n"
        cmd_finish = "if not os.path.exists( \"%s/src_gameXML\" ):\n" % cDir
        cmd_finish += "\tos.mkdir(\"%s/src_gameXML\")\n" % cDir
        cmd_finish += "os.system( \"find . -type f -name '*.gamexml' -exec mv {} %s/src_gameXML/. \\;\" )\n" % cDir
        cL.runSingleJob( cmd_start, cmd_finish )
        cL.endRun()
        if clean:
            cL.clean( acronym )
        jobdb.close()

        os.chdir( "src_gameXML" )
        os.chdir( ".." )

        # add the annotations to the gameXML files
        if verbose > 0:
            print "add annotations to source files..."
            sys.stdout.flush()
        if os.path.exists( "annotations" ):
            os.system( "rm -rf annotations" )
        os.mkdir( "annotations" )
        os.chdir( "annotations" )
        os.system("find ../src_gameXML/ -name '*.gamexml' -exec cp {} . \;")
    
        cDir = os.getcwd()
        if config.get(jobsSectionName, "tmpDir" ) != "":
            tmpDir = config.get(jobsSectionName, "tmpDir")
        else:
            tmpDir = cDir

        groupid = "%s_TEannot_step8_GameXml_addAnnotations" % ( projectName )
        acronym = "GameXmlMaker"
        jobdb = RepetJob( cfgFileName = "%s/%s" % (projectDir, configFileName) )
        cL = Launcher( jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", queue, groupid, acronym )
        cL.beginRun()
        cL.job.jobname = acronym
        cmd_start = ""
        lGXmlFilePaths = glob.glob( "*.gamexml" )
        for gXmlFilePath in lGXmlFilePaths:
            gXmlFileName = os.path.basename( gXmlFilePath )
            prg = "GameXmlMaker.py"
            cmd = prg
            cmd += " -t %s" % os.path.abspath("../%s" % tablesFileName)
            cmd += " -g %s" % gXmlFileName
            cmd += " -v %i" % ( verbose - 1 )
            
            cmd_start += "os.system(\"cp %s .\")\n" % os.path.abspath(gXmlFileName)
            cmd_start += "log = os.system( \""
            cmd_start += cmd
            cmd_start += "\" )\n"
            
        cmd_finish = "os.system( \"find . -type f -name '*.gamexml' -exec mv {} %s/. \\;\" )\n" % cDir
        cL.runSingleJob( cmd_start, cmd_finish )
        cL.endRun()
        if clean:
            cL.clean( acronym )
        jobdb.close()
            
        os.chdir( ".." )
        os.system( "rm -rf src_gameXML/" )
        os.chdir( ".." )
        
    #--------------------------------------------------------------------------
    
    if config.get(sectionName, "drop_tables") == "yes":
        iDb = DbFactory.createInstance()
        lTables = iDb.getTableListFromPattern("%s_%%" % projectName)
        for table in lTables:
            iDb.dropTable(table)
        iDb.close()
            
    if verbose > 0:
        print "step 8 finished successfully\n"
        sys.stdout.flush()
        
#------------------------------------------------------------------------------

if __name__ == "__main__":
    main()
