#!/usr/bin/env python

from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.parsing.FastaParser import FastaParser
import re
import os
from commons.core.LoggerFactory import LoggerFactory

#TODO: change code => always 3 letters, complete with 'X'
DWICKERCODE = {
               "I":"RXX", 
               "II":"DXX", 
               "LTR":"RLX", 
               "DIRS":"RYX", 
               "PLE":"RPX", 
               "LINE":"RIX", 
               "SINE":"RSX", 
               "TIR":"DTX", 
               "Crypton":"DYX", 
               "Helitron":"DHX", 
               "Maverick":"DMX", 
               
               "MITE":"DXX-MITE", 
               "LARD":"RXX-LARD", 
               "TRIM":"RXX-TRIM"
               }

LOG_DEPTH = "repet.tools"

class RenameHeaderClassif(object):
    
    def __init__(self, classifFileName = "", fastaFileName = "", projectName= "", isShorten = False, renameInClassif = False, verbosity = 3):
        self._classifFileName = classifFileName
        self._fastaFileName = fastaFileName
        self._isShorten = isShorten
        self._renameInClassif = renameInClassif
        self._projectName = projectName
        self._outputFileName = ""
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
    
    def setClassifFileName(self, fileName):
        self._classifFileName = fileName
        
    def setFastaFileName(self, fileName):
        self._fastaFileName = fileName
        
    def setOutputFileName(self, fileName = ""):
        if fileName == "":
            fastaFileNameWithoutPath = os.path.basename(self._fastaFileName)
            self._outputFileName = "%s_WickerH.fa" % os.path.splitext(fastaFileNameWithoutPath)[0]
        else:
            self._outputFileName = fileName

    def setShorten(self, isShorten):
        self._isShorten = isShorten
        
    def setRenameInClassif(self, renameInClassif):
        self._renameInClassif = renameInClassif
        
    def setProjectName(self, projectName):
        self._projectName = projectName
        
    def setAttributesFromCmdLine(self):
        examples = "example : "
        parser = RepetOptionParser(description = "Rename headers with Wicker code according to classification", version = "v1.0", epilog = examples)
        parser.add_option( "", "--classif", dest = "classifFileName", help = "classification file (from PASTEC or TEclassifier.py)")
        parser.add_option( "", "--fasta", dest = "fastaFileName", help = "fasta file name")
        parser.add_option( "-s", "--shorten", dest = "shorten", help = "to short the consensus header", action="store_true", default = False)
        parser.add_option( "-r", "--renameInClassif", dest = "renameInClassif", help = "rename the consensus header in the classification file too (<classif file name>_WickerH.classif)", action="store_true", default = False)
        parser.add_option( "-p", "--projectName", dest = "projectName", help = "project name if shorten option (-s)")
        parser.add_option( "-o", "--output", dest = "outputFileName", help = "output file name (default = <fasta file name>_WickerH.fa)", default = "")
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 1)
        options, args = parser.parse_args()
        self._setAttributesFromOptions(options)

    def _setAttributesFromOptions(self, options):
        self.setClassifFileName(options.classifFileName)
        self.setFastaFileName(options.fastaFileName)
        self.setOutputFileName(options.outputFileName)
        self.setShorten(options.shorten)
        self.setRenameInClassif(options.renameInClassif)
        self.setProjectName(options.projectName)
        self._verbosity = options.verbosity

    def _checkAttributes(self):
        if self._classifFileName == "":
            raise Exception("ERROR: No classification file specified in option!")
        elif not FileUtils.isRessourceExists(self._classifFileName):
            raise Exception("ERROR: Can't open classification file : %s" % self._classifFileName)
        if self._fastaFileName == "":
            raise Exception("ERROR: No fasta file specified in option!")
        elif not FileUtils.isRessourceExists(self._fastaFileName):
            raise Exception("ERROR: Can't open fasta file : %s" % self._fastaFileName)
        if self._isShorten == True and self._projectName == "":
            raise Exception("ERROR: With shorten option you need to specify the project name")

    def createClassifInstance(self, classifFileLine):
        lClassifItem = classifFileLine.split("\t")
        consensusName = lClassifItem[0]
        inConfuseness = lClassifItem[3]
        teClass = lClassifItem[4]
        order = lClassifItem[5]
        inCompleteness = lClassifItem[6]
        
        outCompleteness = self._decisionRulesForCompleteness(inCompleteness)
        outConfuseness = self._decisionRulesForConfuseness(inConfuseness)
        code = self._decisionRuleForWickerCode(teClass, order, outConfuseness)
        
        iClassif = Classif(consensusName, code, outConfuseness, outCompleteness, self._projectName, self._isShorten)
        return iClassif
    
    def run(self):
        self._log.info("START of RenameHeaderClassif")
        
        self._checkAttributes()
        fClassif = open(self._classifFileName, "r")
        line = fClassif.readline()
        lClassif = []
        while line:
            lClassif.append(self.createClassifInstance(line))
            line = fClassif.readline()
        fClassif.close()
            
        fOutput = open(self._outputFileName, "w")
        iFastaParser = FastaParser(self._fastaFileName)
        for iSequence in iFastaParser.getIterator():
            for iClassif in lClassif:
                if iClassif.getConsensusName() == iSequence.getName():
                    break
                elif "%s_reversed" % iClassif.getConsensusName() == iSequence.getName():
                    iClassif.setConsensusName(iSequence.getName())
                    break
            iSequence.setName(iClassif.createNewConsensusName())
            fOutput.write(iSequence.printFasta())
        fOutput.close()
        
        if self._renameInClassif:
            self._renameHeadersInClassifFile(lClassif)
        
        self._log.info("END of RenameHeaderClassif")
    
    def _renameHeadersInClassifFile(self, lClassif):
        fileNameWithoutPath = os.path.basename(self._classifFileName)
        outputFileName = "%s_WickerH.classif" % os.path.splitext(fileNameWithoutPath)[0]
        fOut = open(outputFileName, "w")
        f = open(self._classifFileName)
        line = f.readline()
        while line:
            consensusName = line.split("\t")[0]
            for iClassif in lClassif:
                if iClassif.getConsensusName() == consensusName or iClassif.getConsensusName() == "%s_reversed" % consensusName:
                    break
            l = [iClassif.createNewConsensusName()]
            l.extend(line.split("\t")[1:])
            fOut.write("\t".join(l))
            line = f.readline()
        f.close()
        fOut.close()
        
    def _decisionRulesForCompleteness(self, completeness):
        outCompleteness = None
        if completeness == "incomplete":
            outCompleteness = "incomp"
        elif completeness == "complete":
            outCompleteness = "comp"
        elif completeness == "NA":
            outCompleteness = ""
        return outCompleteness

    def _decisionRulesForConfuseness(self, inConfuseness):
        outConfuseness = None
        if inConfuseness == "ok":
            outConfuseness = ""
        elif inConfuseness == "PotentialChimeric":
            outConfuseness = "chim"
        return outConfuseness

    def _decisionRuleForWickerCode(self, teClass, order, confusness):
        code = None
        if order in DWICKERCODE.keys():
            code = DWICKERCODE[order]
        elif order == "noCat" and teClass == "noCat":
            code = "noCat"
        elif re.search("\|",order) and teClass == "noCat":
            code = "XXX"
        elif re.search("\|",order) and teClass == "NA":
            code = "nonTE"
        elif order == "noCat" or re.search("\|",order) :
            code = DWICKERCODE[teClass]
        elif order == "SSR":
            code = "SSR"
        elif order == "rDNA":
            code = "rDNA"
        elif order == "PotentialHostGene":
            code = "PotentialHostGene"
        return code

class Classif(object):
    
    def __init__(self, consensusName, code, outConfuseness, outCompleteness, projectName = "", isShorten = False):
        self._consensusName = consensusName
        self._code = code
        self._confusness = outConfuseness
        self._completeness = outCompleteness
        self._projectName = projectName
        self._isShorten = isShorten
        
    def __eq__(self, o):
        return self._consensusName == o._consensusName and self._code == o._code \
            and self._confusness == o._confusness and self._completeness == o._completeness
    
    def createNewConsensusName(self):
        pastecClassif = "%s" % self._code
        if self._completeness != "":
            pastecClassif += "-%s" % self._completeness
        if self._confusness != "":
            pastecClassif += "-%s" % self._confusness
            
        if self._isShorten:
            pattern = "%s_[a-zA-Z0-9]+_[a-zA-Z0-9]+_[a-zA-Z0-9_]+" % ( self._projectName ) 
            if re.match(pattern, self._consensusName):
                header = self.shortenConsensusName()
                header = "%s_%s" % (pastecClassif, header)
            else:
                header = "%s_%s" % (pastecClassif, self._consensusName)
        else:
            header = "%s_%s" % (pastecClassif, self._consensusName)
            
        return header
    
    def shortenConsensusName(self):
        desc = self._consensusName.split(self._projectName)[1]
        palignMeth = desc.split("_")[1]
        clustMeth = desc.split("_")[2]
        clustID = desc.split("_")[3]
        lmalignMeth = desc.split("_")[4:]
        if len(lmalignMeth) > 2:
            malignMeth = "%s%s_%s" % (lmalignMeth[0],lmalignMeth[1],lmalignMeth[2])
        else:
            malignMeth = "".join(lmalignMeth) 
        consensusShorten = "%s-%s-%s%s-%s" % ( self._projectName, palignMeth[0], clustMeth[0], clustID, malignMeth ) 
        return consensusShorten
    
    def getConsensusName(self):
        return self._consensusName
    
    def setConsensusName(self, consensusName):
        self._consensusName = consensusName
    
    def setProjectName(self, projectName):
        self._projectName = projectName
        
if __name__ == "__main__":
    iRHC = RenameHeaderClassif()
    iRHC.setAttributesFromCmdLine()
    iRHC.run()