#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Pipeline for the identification of satellites in genomic sequences.
"""

import os
import sys
import time
import glob
import shutil

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append(os.environ["REPET_PATH"])
if not "PYTHONPATH" in os.environ.keys():
    os.environ["PYTHONPATH"] = os.environ["REPET_PATH"]
else:
    os.environ["PYTHONPATH"] = "%s:%s" % (os.environ["REPET_PATH"], os.environ["PYTHONPATH"])

from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.checker.ConfigChecker import ConfigRules, ConfigChecker
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory
from commons.core.launcher.Launcher import Launcher
from commons.core.launcher.LauncherUtils import LauncherUtils
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.coord.MapUtils import MapUtils
from commons.core.coord.ConvCoord import ConvCoord
from commons.core.coord.Path import Path
from commons.core.coord.PathUtils import PathUtils
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.LoggerFactory import LoggerFactory
from commons.tools.GFF3Maker import GFF3Maker
from commons.tools.RmvPairAlignInChunkOverlaps import RmvPairAlignInChunkOverlaps
from commons.tools.PrepareBatches import PrepareBatches
from commons.tools.ShortenConsensusHeader import ShortenConsensusHeader
from satannot_pipe.DetectTandemByCluster import DetectTandemByCluster

LOG_DEPTH = "repet.SATannot"

class SATannotPipeline(object):
    """
    This program is a pipeline for the identification of satellites in genomic sequences.
    """
    
    def __init__(self, smpAlign="Pals",clustHsp="Piler", multAlign="MAP", configFileName = "SATannot.cfg", step = 1, verbosity = 1):
        self._pipelineName = "SATannot"
        self._smpAlign = smpAlign
        self._clustHsp = clustHsp
        self._multAlign = multAlign
        
        self._configFileName = configFileName
        self._step = step
        self._verbosity = verbosity
        self._iConfig = None
        self._projectName = ""
        self._projectDir = ""
        self._inFileName = ""
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
  
    def setAttributesFromCmdLine(self):
        description = "Launch pipeline for the identification of satellites in genomic sequences."
        epilog= "\n"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-s", "--smplAlign",    dest = "smplAlign", action = "store",       type = "string", help = "program to self-align (Pals)",   default = "Pals")
        parser.add_option("-c", "--clustHsp",    dest = "clustHsp", action = "store",       type = "string", help = "program to cluster HSPs (Piler)",   default = "Piler")
        parser.add_option("-m", "--multAlign",    dest = "multAlign", action = "store",       type = "string", help = "program to build MSAs  (MAP/Mafft/Muscle/Clustalw/Tcoffee/Prank)",   default = "MAP")
        #parser.add_option("-f", "--filter",    dest = "filter", action = "store",       type = "string", help = "program to select one pyramid" default="1")
        parser.add_option("-C", "--config",    dest = "configFileName", action = "store",       type = "string", help = "input config file name [optional] [format: cfg]",   default = "SATannot.cfg")
        parser.add_option("-S", "--step",      dest = "step",           action="store",         type="int",      help = "step (1/2/3/4/5/6)",                                  default = 1)
        parser.add_option("-v", "--verbosity", dest = "verbosity",      action = "store",       type = "int",    help = "verbosity [optional] [default: 1]",                 default = 1)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._configFileName = options.configFileName
        self._inFileName = "%s.fa" % self._projectName 
        self._step = options.step
        self._verbosity = options.verbosity
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def _checkOptions(self):
        if self._step not in [1,2,3,4,5,6]:
            self._logAndRaise("ERROR: Bad step '%i'. Must be 1,2,3,4,5 or 6" % self._step)
            
    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section="repet_env", mandatory=True)
        changelogVersion = CheckerUtils.getVersionFromChangelogFile("%s/satannot_pipe/doc/CHANGELOG" % os.environ["REPET_PATH"])
        iConfigRules.addRuleOption(section="repet_env", option ="repet_version", set =(changelogVersion), mandatory=True)
        iConfigRules.addRuleOption(section="repet_env", option ="repet_host", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_user", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_pw", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_db", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_port", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_job_manager", mandatory=True, type="string")
        
        iConfigRules.addRuleSection(section="project", mandatory=True)
        iConfigRules.addRuleOption(section="project", option ="project_name", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="project", option ="project_dir", mandatory=True, type="string")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._projectName = self._iConfig.get("project", "project_name")
        self._projectDir = self._iConfig.get("project", "project_dir")
        
        if not CheckerUtils.isMax15Char(self._projectName):
            self._logAndRaise("ERROR: project name must have 15 character max")
        if not CheckerUtils.isCharAlphanumOrUnderscore(self._projectName):
            self._logAndRaise("ERROR: project name must contain only alphanumeric or underscore character")
        if not os.path.exists(self._projectDir):
            self._logAndRaise("ERROR: project folder does not exist")
          
    def _setupEnv(self):
        os.environ["REPET_HOST"] = self._iConfig.get("repet_env", "repet_host")
        os.environ["REPET_USER"] = self._iConfig.get("repet_env", "repet_user")
        os.environ["REPET_PW"] = self._iConfig.get("repet_env", "repet_pw")
        os.environ["REPET_DB"] = self._iConfig.get("repet_env", "repet_db")
        os.environ["REPET_PORT"] = self._iConfig.get("repet_env", "repet_port")
        os.environ["REPET_JOB_MANAGER"] = self._iConfig.get("repet_env", "repet_job_manager")
        os.environ["REPET_QUEUE"] = self._iConfig.get("repet_env", "repet_job_manager")
        os.environ["REPET_JOBS"] = "MySQL"
        
    def run(self):
        self._log.info("START %s (%s)" % (self._pipelineName, time.strftime("%Y-%m-%d %H:%M:%S")))
        self._checkOptions()
        self._checkConfig()
        self._setupEnv()
        
        if self._step == 1:
            iPrepareBatches = PrepareBatches(self._pipelineName, self._projectDir, self._projectName, self._iConfig, self._verbosity)
            iPrepareBatches.run()
        elif self._step == 2:
            iSelfAlignAndConcatenate = SelfAlignAndConcatenate(self._pipelineName, self._configFileName, "self_align", self._verbosity)
            iSelfAlignAndConcatenate.run()
        elif self._step == 3:
            iClusterHsp = ClusterHsp(self._pipelineName, self._configFileName, "cluster_HSPs", self._verbosity)
            iClusterHsp.run()
        elif self._step == 4:
            iBuildConsensus = BuildConsensus(self._pipelineName, self._configFileName, "build_consensus", self._verbosity)
            iBuildConsensus.run()
        elif self._step == 5:
            iDetectTandem = DetectTandem(self._pipelineName, self._configFileName, "detect_tandem", self._verbosity)
            iDetectTandem.run()
        elif self._step == 6:
            iBuildTandemArrays = BuildTandemArrays(self._pipelineName, self._configFileName, "", self._verbosity)
            iBuildTandemArrays.run()
            
        self._log.info("END %s (%s)" % (self._pipelineName, time.strftime("%Y-%m-%d %H:%M:%S")))

            
class PipelineStep(object):
    
    def __init__(self, configFileName, configSection, verbosity = 1):
        self._configFileName = configFileName
        self._configSection = configSection
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)

    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)


class SelfAlignAndConcatenate(PipelineStep):
    
    def __init__(self, pipelineName, configFileName, configSection, verbosity = 1):
        """
        self-alignment of the input data bank
    
        @param smplAlign: name of the alignment program (Pals)
        @type smplAlign: string
        """
        PipelineStep.__init__(self, configFileName, configSection, verbosity)
        self._pipelineName = pipelineName
        self._catFiles = True
        self._allByAll = False
    
    def _setAttributesFromConfig(self):
        iConfigRules = ConfigRules()
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._projectName = iConfig.get("project", "project_name")
        self._projectDir = iConfig.get("project", "project_dir")
        
        self._smplAlign = iConfig.get(self._configSection,"smplAlign")
        
        self._identity = iConfig.get(self._configSection,"identity")
        self._length = iConfig.get(self._configSection,"length")
        
        self._minEvalue = iConfig.get(self._configSection,"min_Evalue")
        self._minIdentity = iConfig.get(self._configSection,"min_identity")
        self._minLength = iConfig.get(self._configSection,"min_length")
        self._maxLength = iConfig.get(self._configSection,"max_length")
        
        self._chunkLength = iConfig.get("prepare_batches","chunk_length")
        self._chunkOverlap = iConfig.get("prepare_batches","chunk_overlap")
        
        self._workdir = "%s/%s_%s" % (self._projectDir, self._projectName,self._smplAlign)
        
        self._tmpDir = iConfig.get(self._configSection, "tmpDir")
        self._queue = iConfig.get(self._configSection,"resources")
        self._filterHSP = True if iConfig.get(self._configSection, "filter_HSP") == "yes" else False
        self._doClean = True if iConfig.get(self._configSection, "clean") == "yes" else False
    
    def getPalsCmd(self, ifile, iLauncher):
        lArgs = []
        lArgs.append("-q %s" % (ifile ))
        lArgs.append("-s %s/%s_%s/%s_chunks.fa" % (self._projectDir,self._projectName, "db", self._projectName))
        lArgs.append("--length %s" % self._length) #400
        lArgs.append("--id %s" % self._identity) #94
        return iLauncher.getSystemCommand("LaunchPals.py", lArgs) 
    
    def concatenateFiles(self, groupid, sameSequences ):
        self._log.info("concatenate the results of each job")
    
        if os.path.exists( "%s.gff" % ( groupid ) ):
            os.remove( "%s.gff" % ( groupid ) )
    
        lGffFiles = glob.glob( "*_Pals.gff" )
        lGffFiles.sort()
        for gffFile in lGffFiles:
            prg = "cat"
            cmd = "%s %s >> %s.gff" % (prg, gffFile, groupid )
            log = os.system( cmd )
            if log != 0:
                self._logAndRaise("Error: %s returned %i for file : %s" % ( prg, log, gffFile ))
            if self._doClean:
                prg = "rm"
                cmd = "%s -f %s" % (prg, gffFile)
                log = os.system( cmd )
                if log != 0:
                    self._logAndRaise("Error: %s returned %i" % ( prg, log ))
    
        self._log.info("finished concatenating the results of each job")
        
        self._log.info("converting pals alignment to .align")
        if os.path.exists( "%s.align" % ( groupid ) ):
            os.remove( "%s.align" % ( groupid ) )
    
        prg = "pals2align.py"
        cmd = prg
        cmd += " -i %s.gff" % ( groupid )
        if sameSequences:
            cmd += " -a"
        cmd += " -o %s.align_tmp" % ( groupid )
        log = os.system( cmd )
        if log != 0:
            self._logAndRaise("Error: %s returned %i" % ( prg, log ))
    
        self._log.info("finished converting pals alignment to .align")
        
      
        os.system( "mv %s.align_tmp %s.align" % ( groupid, groupid ) ) 
            
        if self._doClean:
            os.remove("%s.gff" % groupid)
    
    def removeChunkOverlaps(self):
        # remove the chunk overlaps
        self._log.info("Removing chunk overlaps")
        rpaico = RmvPairAlignInChunkOverlaps()
        rpaico.setInputFileName( "%s.align" % (self._projectName ) )
        rpaico.setChunkLength(self._chunkLength)
        rpaico.setChunkOverlap(self._chunkOverlap)
        rpaico.setMargin( 10 )
        rpaico.setOutputFileName( "%s.align.not_over" % (self._projectName ) )
        rpaico.setVerbosityLevel(self._verbosity - 2 )
        rpaico.run()
        self._log.info("Finished removing chunk overlaps")
    
    def filterHSP(self):
        self._log.info("Filtering HSPs")
        cmd = "FilterAlign.py"
        cmd += " -i %s.align.not_over" % self._projectName
        cmd += " -E %s" % self._minEvalue
        cmd += " -I %s" % self._minIdentity
        cmd += " -l %s" % self._minLength
        cmd += " -L %s" % self._maxLength
        cmd += " -v %i" % (self._verbosity - 2)
        log = os.system(cmd)
        if log != 0:
            self._logAndRaise("Error: %s returned '%i'" % (cmd, log))
        self._log.info("Finished HSPs filtering")
    
    def run(self):
        self._setAttributesFromConfig()
        if not os.path.exists(self._workdir):
            os.makedirs(self._workdir)
        os.chdir(self._workdir)
        
        # launch Pals on all the chunks in parallel
        cDir = os.getcwd()
        if self._tmpDir != "":
            tmpDir = self._tmpDir
        else:
            tmpDir = cDir
            
        groupid = "%s_SATannot_%s" % (self._projectName, self._smplAlign)
        acronym = self._smplAlign
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", self._queue, groupid)
        
        lCmdsTuples = []
        lFiles = FileUtils.getFileNamesList("%s/%s_db/batches" % (self._projectDir, self._projectName), "batch*")
        
        for ifile in lFiles:
            lCmds = []
            lCmds.append(self.getPalsCmd(ifile, iLauncher))
            lCmdStart = []
            lCmdStart.append("shutil.copy('%s/%s_db/batches/%s', '.' )" % (self._projectDir, self._projectName, ifile))
            
            lCmdFinish = []
            lCmdFinish.append("shutil.move('%s_%s.gff', '%s')" % (os.path.splitext(ifile)[0], self._smplAlign, cDir))
            
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)  
        
        # concatenate all individual files in a big one
        if self._catFiles:
            self.concatenateFiles(groupid, self._allByAll)
    
        os.system( "mv %s_SATannot_%s.align %s.align" % (self._projectName,  self._smplAlign,self._projectName ) )
    
        self.removeChunkOverlaps()
        if self._filterHSP:
            self.filterHSP()
    
        # clean
        if self._doClean:
            # for each job, remove the file containing the HSPs
            cmd = "find . -name \"batch_*.align\" -exec rm {} \;"
            log = os.system( cmd )
            if log != 0:
                self._logAndRaise("Error while removing the 'align' files from the jobs")
    
            # remove the files "_cut"
            cmd = "find . -name \"%s*_cut*\" -exec rm {} \;" % (self._projectName )
            log = os.system( cmd )
            if log != 0:
                self._logAndRaise("Error while removing the '_cut' files")
    
            # remove the files "Nstretch.map"
            cmd = "find . -name \"*Nstretch.map\" -exec rm {} \;"
            log = os.system( cmd )
            if log != 0:
                self._logAndRaise("Error while removing the 'Nstretch.map' files")
    
            # remove the temporary file(s) with the HSPs
            if self._filterHSP:
                cmd = "rm -f %s.align %s.align.not_over" % (self._projectName,self._projectName )
                log = os.system( cmd )
                if log != 0:
                    self._logAndRaise("Error while removing an intermediary file")
    
            # remove the genomic sequences (symlink and batches)
            os.system( "rm -rf %s.fa batches" % (self._projectName ) )
    
        # clean the 'map' file
        if self._doClean and self._filterHSP: 
            cmd = "rm -f %s.align.not_over.filtered.map" % (self._projectName )
            log = os.system( cmd )
            if log != 0:
                self._logAndRaise("Error while removing an intermediary file")
    
        os.chdir( ".." )
        self._log.info("SelfAlignAndConcatenate finished successfully")


class ClusterHsp(PipelineStep):
    
    def __init__(self, pipelineName, configFileName, configSection, verbosity = 1):
        """
        step 3: cluster the resulting HSPs (high-scoring segment pairs)
        @param smplAlign: name of the alignment program (Pals)
        @type smplAlign: string
    
        @param clustHsp: name of the clustering program (Piler)
        @type clustHsp: string
        """
        PipelineStep.__init__(self, configFileName, configSection, verbosity)
        self._pipelineName = pipelineName
        
    def _setAttributesFromConfig(self):
        iConfigRules = ConfigRules()
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._projectName = iConfig.get("project", "project_name")
        self._projectDir = iConfig.get("project", "project_dir")
        self._smplAlign = iConfig.get("self_align", "smplAlign")
        self._clustHsp = iConfig.get(self._configSection, "clustHsp")
        self._minSeqPerGroup = iConfig.get(self._configSection, "minNbSeqPerGroup")
        self._nbLongestSeqPerGroup = iConfig.get(self._configSection,"nbLongestSeqPerGroup")
        self._maxSeqLength =  iConfig.get(self._configSection,"maxSeqLength")
        
        self._filterHsp = True if iConfig.get("self_align","filter_HSP") == "yes" else False
        
        self._doClean = True if iConfig.get(self._configSection, "clean") == "yes" else False
        
        self._workdir = "%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp)
        
        self._tmpDir = iConfig.get(self._configSection, "tmpDir")
        self._queue = iConfig.get(self._configSection,"resources")
        
    def checkConfig(self):
        if os.path.exists(self._workdir):
            self._logAndRaise("Error: directory %s already exists" % self._workdir)
    
    def getFilterSeqClusterCmd(self, dirName, minSeq, maxSeq, sectionNameClusterHSP, verbose, logFileName, iLauncher):
        lArgs = []
        lArgs.append("-i %s.map" % dirName)
        lArgs.append("-m %s" % minSeq)
        lArgs.append("-M %s" % maxSeq)
        lArgs.append("-L %s" % self._maxSeqLength)
        lArgs.append("-d")
        lArgs.append("-v 3 > %s" % logFileName)
        return iLauncher.getSystemCommand("FilterSeqClusters.py", lArgs) 
    
    def getLaunchReconOrPilerCmd(self, clustHsp, verbosity, alignFileName, bankName, iLauncher, doClean):
        lArgs = []
        lArgs.append("-p %s_%s" % (self._projectName, self._smplAlign))
        lArgs.append("-f %s" % bankName)
        lArgs.append("-a %s" % alignFileName)
        if doClean:
            lArgs.append("-c")
        lArgs.append("-v %i" % (verbosity - 1))
        return iLauncher.getSystemCommand("launch%s.py" % (clustHsp), lArgs) 
    

    def run(self):
        self._setAttributesFromConfig()
        if not os.path.exists(self._workdir):
            os.makedirs(self._workdir)
        os.chdir(self._workdir)
        
        self._log.info("beginning of step 3 self-alignment with %s clustering with %s " % (self._smplAlign, self._clustHsp)) 

        alignFileName = self._projectName + ".align.not_over"
        
        if self._filterHsp :
            alignFileName += ".filtered"
        pathToAlignFile = "%s/%s_%s/%s" % (self._projectDir, self._projectName, self._smplAlign, alignFileName)
    
        bankName = "%s_chunks.fa" % self._projectName
        pathToBank = "%s/%s_db/%s" % (self._projectDir, self._projectName, bankName)
        imageFileName = "%s_%s_Piler-trs_images.gff" % (self._projectName, self._smplAlign )
        
        ##############################################"
        cDir = os.getcwd()
        if self._tmpDir != "":
            tmpDir = self._tmpDir
        else:
            tmpDir = cDir
            
        groupid = "%s_SATannot_%s" % (self._projectName, self._clustHsp)
        acronym = self._clustHsp
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(jobdb, os.getcwd(), "", "", cDir, tmpDir, "jobs", self._queue, groupid)
        
        self._log.info("Running Piler HSP clustering")
        lCmdsTuples = []
        
        lCmds = []
        lCmdStart = []
        lCmdFinish = []
        
        cmd = self.getLaunchReconOrPilerCmd(self._clustHsp, self._verbosity, alignFileName, bankName, iLauncher, False)
        lCmds.append(cmd)
        
        cmd = self.getLaunchReconOrPilerCmd("PilerTA", self._verbosity, imageFileName, bankName, iLauncher, self._doClean)
        lCmds.append(cmd)
        
        lCmdStart.append("os.symlink(\"%s\", \"%s\")" % (pathToBank, bankName))
        lCmdStart.append("os.symlink(\"%s\", \"%s\")" % (pathToAlignFile, alignFileName))
        
        lCmdFinish.append("shutil.move(\"%s_%s_Piler.map\", \"%s/.\")" % (self._projectName, self._smplAlign, cDir))
        lCmdFinish.append("shutil.move(\"%s_%s_Piler.fa\", \"%s/.\")" % (self._projectName, self._smplAlign, cDir))
        lCmdFinish.append("shutil.move(\"%s_%s_Piler-tan.fa\", \"%s/.\")" % (self._projectName, self._smplAlign, cDir))
        lCmdFinish.append("shutil.move(\"%s.gff\", \"%s/.\")" % (alignFileName, cDir))
        lCmdFinish.append("shutil.move(\"%s_%s_Piler-trs.gff\", \"%s/.\")" % (self._projectName, self._smplAlign, cDir))
        lCmdFinish.append("shutil.move(\"%s\", \"%s/.\")" % (imageFileName, cDir))
        
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)  
        
        self._log.info("Finished running Piler HSP clustering")
        
        #TODO: FilterSeqClusters.py to limit sequences number to 20 in a cluster
#        cmd = self.getFilterSeqClusterCmd(dirName, minSeq, maxSeq, sectionName, verbose, logFileName, iLauncher)
#        lCmds.append(cmd)
        
        os.chdir( ".." )
        self._log.info("step 3 finished successfully")


class BuildConsensus(PipelineStep):
    
    def __init__(self, pipelineName, configFileName, configSection, verbosity = 1):
        """
        step 4: build consensus from multiple alignments
    
        @param smplAlign: name of the alignment program (Pals)
        @type smplAlign: string
    
        @param clustHsp: name of the clustering program (Piler)
        @type clustHsp: string
    
        @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
        @type multAlign: string
        """
        PipelineStep.__init__(self, configFileName, configSection, verbosity)
        self._pipelineName = pipelineName
        
        #old params smplAlign, clustHsp, multAlign
        
    def _setAttributesFromConfig(self):
        iConfigRules = ConfigRules()
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._projectName = iConfig.get("project", "project_name")
        self._projectDir = iConfig.get("project", "project_dir")
        
        self._smplAlign = iConfig.get("self_align", "smplAlign")
        self._filterHsp = True if iConfig.get("self_align", "filter_HSP") == "yes" else False
        
        self._clustHsp = iConfig.get("cluster_HSPs", "clustHsp")
        
        self._multAlign = iConfig.get(self._configSection, "multAlign") 
        self._minBasesPerSite = iConfig.get(self._configSection,"minBasesPerSite")
        
        self._tmpDir = iConfig.get(self._configSection, "tmpDir")
        self._queue = iConfig.get(self._configSection,"resources")
        self._doClean = True if iConfig.get(self._configSection, "clean") == "yes" else False
        
        self._workdir = "%s_%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp, self._multAlign)
        
    def checkConfig(self):
        if os.path.exists(self._workdir):
            self._logAndRaise("Error: directory %s already exists" % self._workdir)
    
    def getLaunchMapCmd(self, inFile, iLauncher):
        lArgs = []
        lArgs.append("-i %s" % inFile)
        if self._doClean:
            lArgs.append("-c")
        lArgs.append("-v %i" % self._verbosity)
        return iLauncher.getSystemCommand("LaunchMap.py", lArgs)
    
    def getMafftProgramLauncherCmd(self, inFile, iLauncher):
        lArgs = []
        lArgs.append("-i %s" % inFile)
        if self._doClean:
            lArgs.append("-c")
        lArgs.append("-v %i" % self._verbosity)
        return iLauncher.getSystemCommand("MafftProgramLauncher.py", lArgs)

    def getBuildConsLibCmd(self, prefix, multAlign, pyramidHeader, iLauncher):
        lArgs = []
        lArgs.append("-d %s" % os.getcwd())
        lArgs.append("-r \'*.fa_aln\'")
        lArgs.append("-n %s" % self._minBasesPerSite)
        lArgs.append("-o %s_%s_consensus_long.fa" % (prefix, multAlign))#consensus_long.fa
        if pyramidHeader:
            lArgs.append("-H")
        if self._doClean:
            lArgs.append("-c")
        lArgs.append("-v %i" % (self._verbosity - 1))
        return iLauncher.getSystemCommand("buildConsLib.py", lArgs)
    
        
    def run(self):
        self._setAttributesFromConfig()
        if not os.path.exists(self._workdir):
            os.makedirs(self._workdir)
        os.chdir(self._workdir)
        self._log.info("beginning of step 4 self-alignment with %s clustering with %s multiple alignment with %s" % (self._smplAlign, self._clustHsp, self._multAlign)) 

        pilerTanFileBase = "%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp)
        suffix = "-tan.fa"
        pilerTanFileName = pilerTanFileBase +  suffix
        
        pilerTanPath =  "%s/%s/%s" % ( self._projectDir, pilerTanFileBase, pilerTanFileName )
        if not os.path.exists(pilerTanPath):
            self._logAndRaise("Error: %s not found" % ( pilerTanPath))
        os.symlink(pilerTanPath, pilerTanFileName)
    
        if os.path.exists( "%s_%s" % (pilerTanFileBase, self._multAlign)):
            self._logAndRaise("Error: directory %s_%s/ already exists" % ( pilerTanFileBase, self._multAlign ))

        
        # split each cluster into a specific fasta file
        self._log.info("Splitting sequences per cluster")
        FastaUtils.splitSeqPerCluster(inFileName=pilerTanFileName, clusteringMethod=self._clustHsp,simplifyHeader=False, createDir=False, verbose = self._verbosity-1 )
        self._log.info("Finished Splitting sequences per cluster")
        
        self._log.info("Building multiple sequence alignments (MSA) for each file")
        # build multiple sequence alignments (MSA) for each file
        #TODO: refactor common code , cf TEdenovo line 1405
        if self._multAlign in [ "Map", "MAP" ] and not CheckerUtils.isExecutableInUserPath("rpt_map") :
            self._logAndRaise("ERROR: map must be in your path")
        if self._multAlign == "Mafft" and not CheckerUtils.isExecutableInUserPath("mafft") :
            self._logAndRaise("ERROR: mafft must be in your path")
        
        lFiles = glob.glob("seqCluster*.fa")
        lFileSizeTuples = []
        for fileName in lFiles:
            iBioseqDB = BioseqDB(fileName)
            if iBioseqDB.getSize() == 1:
                iBioseqDB.upCase()
                iBioseqDB.save("%s.fa_aln" % fileName)
            else:
                fileSize = os.path.getsize(fileName)
                lFileSizeTuples.append((fileName, fileSize))
        lFileSizeList = LauncherUtils.createHomogeneousSizeList(lFileSizeTuples, 100000)
        
        inFileDirectory = os.getcwd()
        groupid = "%s_SATannot_%s_%s_%s" % (self._projectName, self._smplAlign,  self._clustHsp, self._multAlign)
        acronym = self._multAlign
        
        cDir = os.getcwd()
        if self._tmpDir != "":
            tmpDir = self._tmpDir
        else:
            tmpDir = cDir
        
        iDb = DbFactory.createInstance()
        iTJA = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(iTJA, os.getcwd(), "", "", cDir, tmpDir, "jobs", self._queue, groupid)
        lCmdsTuples = []
        
        for lFiles in lFileSizeList:
            lCmds = []
            lCmdStart = []
            lCmdFinish = []
            for ifile in lFiles:
                if self._multAlign in [ "Map", "MAP" ]:
                    lCmds.append(self.getLaunchMapCmd(ifile, iLauncher))
                elif self._multAlign == "Mafft":
                    #TODO: test Mafft in functional test
                    lCmds.append(self.getMafftProgramLauncherCmd(ifile, iLauncher))
                lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (inFileDirectory, ifile, ifile))
                lCmdFinish.append("shutil.copy(\"%s.fa_aln\", \"%s/.\")" % (ifile, inFileDirectory))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)  
        
        self._log.info("Finished Building multiple sequence alignments (MSA) for each file")
    
    
        self._log.info("Building the consensus for each MSA and concatenate them into a single fasta file")
        # build the consensus for each MSA and concatenate them into a single fasta file
        groupid = "%s_%s_%s_buildConsLib" % (self._pipelineName, self._projectName, self._clustHsp)
        
        iDb = DbFactory.createInstance()
        iTJA = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(iTJA, os.getcwd(), "", "", cDir, tmpDir, "jobs", self._queue, groupid)
        
        lCmds = []
        lCmdStart = []
        lCmds.append(self.getBuildConsLibCmd(pilerTanFileBase, self._multAlign, True, iLauncher))
        
        lFiles = glob.glob(cDir + "/*.fa_aln")
        
        for ifile in lFiles:
            lCmdStart.append("os.symlink(\"%s\", os.getcwd() + \"/%s\")" % (ifile, os.path.basename(ifile)))
        lCmdsTuples = []
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart))
        iLauncher.runLauncherForMultipleJobs("BuildConsLib_%s" % self._multAlign, lCmdsTuples, self._doClean)
            
        if not os.path.exists( "%s_%s_consensus_long.fa" % ( pilerTanFileBase, self._multAlign)  ):
            self._logAndRaise("Error: %s_%s_consensus_long.fa not in project directory" % ( pilerTanFileBase, self._multAlign  ))
        self._log.info("Finished Building the consensus for each MSA and concatenate them into a single fasta file")
            
        #shorten Header from consensus
        inFileName = "%s_%s_consensus_long.fa" % ( pilerTanFileBase, self._multAlign )
        outFileName = "%s_%s_consensus.fa" % ( pilerTanFileBase, self._multAlign )
        
        iShortenConsensusHeader = ShortenConsensusHeader(fastaFileName = inFileName, outFileName = outFileName, \
        projectName = self._projectName , selfAlign = self._smplAlign, clustering=self._clustHsp,\
        multAlign= self._multAlign , pyramidHeader=True, verbosity = self._verbosity - 1 )
        
        iShortenConsensusHeader.run()
        
        os.chdir( ".." )
        # clean
        if self._doClean:
            os.system( "find . -name \"profil*\" -exec rm {} \;" )
            os.system( "find . -name \"seqCluster*\" -exec rm {} \;" )
    
        self._log.info("step 4 finished successfully")



class DetectTandem(PipelineStep):
    
    def __init__(self, pipelineName, configFileName, configSection, verbosity = 1):
        """
        step 5: detect the divergent units
    
        @param smplAlign: name of the alignment program (Blaster or Pals)
        @type smplAlign: string
    
        @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
        @type clustHsp: string
    
        @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
        @type multAlign: string
        """
      
        PipelineStep.__init__(self, configFileName, configSection, verbosity)
        self._pipelineName = pipelineName
        
    def _setAttributesFromConfig(self):
        iConfigRules = ConfigRules()
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._projectName = iConfig.get("project", "project_name")
        self._projectDir = iConfig.get("project", "project_dir")
        
        self._smplAlign = iConfig.get("self_align", "smplAlign")
        self._clustHsp = iConfig.get("cluster_HSPs", "clustHsp")
        self._multAlign = iConfig.get("build_consensus", "multAlign") 
        
        self._BLR_sensitivity = iConfig.get(self._configSection, "BLR_sensitivity")
        self._Blast = iConfig.get(self._configSection, "blast")
        
        self._tmpDir = iConfig.get(self._configSection, "tmpDir")
        self._queue = iConfig.get(self._configSection, "resources")
        self._doClean = True if iConfig.get(self._configSection, "clean") == "yes" else False
        
        self._workdir = "%s_%s_%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp, self._multAlign, "Blaster_Matcher")
        
    def checkConfig(self):
        if os.path.exists(self._workdir):
            self._logAndRaise("Error: directory %s already exists" % self._workdir)

    def removeInPathFileDuplicatedPathOnQueryNameQueryCoordAndSubjectName(self, inputFile, outputFile):
        f = open(inputFile, "r")
        line = f.readline()
        previousQuery = ""
        previousCluster = ""
        lPaths = []
        while line:
            iPath = Path()
            iPath.setFromString(line)
            query = iPath.getQueryName()
            subject = iPath.getSubjectName()
            cluster = int(subject.split("Piler_")[1].split("_")[0])
            if (query != previousQuery or cluster != previousCluster) and lPaths != []: 
                lPathsWithoutDuplicate = PathUtils.getPathListWithoutDuplicatesOnQueryCoord(lPaths)
                PathUtils.writeListInFile(lPathsWithoutDuplicate, outputFile, "a")
                lPaths = []
            lPaths.append(iPath)
            previousQuery = query
            previousCluster = cluster
            line = f.readline()
        lPathsWithoutDuplicate = PathUtils.getPathListWithoutDuplicatesOnQueryCoord(lPaths)
        PathUtils.writeListInFile(lPathsWithoutDuplicate, outputFile, "a")
        f.close()
    
    def writeBlasterInParallelConfig(self, configFileName):
        with open(configFileName, "w") as fh:
            fh.write("[jobs]\n")
            fh.write("resources: %s\n" % self._queue)
            fh.write("tmpDir: %s\n" % self._tmpDir)
            fh.write("copy: %s\n" % self._copyOnNode)
            if self._doClean:
                fh.write("clean: yes\n")
            else:
                fh.write("clean: no\n")
            fh.write("\n")
            fh.write("[prepare_data]\n")
            fh.write("chunk_length: 0\n")
            fh.write("chunk_overlap: 0\n")
            fh.write("\n")
            fh.write("[alignment]\n")
            fh.write("blast: %s\n" % self._Blast)
            fh.write("Evalue: 1e-10\n")
            fh.write("length: 20\n")
            fh.write("identity: 0\n")
    
    def getMatcherCmd(self, iLauncher, ifile):
#       matcher with all default settings
        lArgs = []
        lArgs.append("-m %s" % ifile)
#        lArgs.append("-j -x")
        lArgs.append("-v %s" % self._verbosity)
        
        return iLauncher.getSystemCommand("matcher", lArgs)

    def run(self):
        self._setAttributesFromConfig()
        if not os.path.exists(self._workdir):
            os.makedirs(self._workdir)
        os.chdir(self._workdir)
        os.symlink("../%s" % self._configFileName, self._configFileName)
        
        self._log.info("beginning of step 5 self-alignment with PALS clustering with %s multiple alignment with %s mapping consensus with %s" % (self._clustHsp, self._multAlign, self._smplAlign))
        self._log.debug("detect the divergent units")
        
        path2chr2chunkMapFile = "%s/%s_db/%s_chunks.map" % (self._projectDir, self._projectName, self._projectName)
        path2chr2chunkFastaFile = "%s/%s_db/%s_chunks.fa" % (self._projectDir, self._projectName, self._projectName)
    
        if not os.path.exists( path2chr2chunkFastaFile ):
            self._logAndRaise("Error: %s not in project directory" % ( path2chr2chunkFastaFile ))
        else:
            os.system ("ln -s %s ." % ( path2chr2chunkFastaFile ) )
    
        #launch Blaster on all the batches in parallel
        #TODO: refactor, same as teannot s6 line 1536
        prefix = "%s_%s_%s_%s" % (self._projectName, self._smplAlign , self._clustHsp, self._multAlign )
        consensusBaseName = "%s_%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp, self._multAlign )
        tandemCons = "%s_consensus.fa" % ( consensusBaseName )
        bankName = "%s/%s/%s"% (self._projectDir, prefix, tandemCons )
        batchesDir = "%s/%s_db/batches" % (self._projectDir, self._projectName)
        batchPattern = "batch_"
        typeBlast = "blastn" #MAY BE should it be blastn, blastx, tblastx or something else?
        self._copyOnNode = "no" #??? what to use
        tandemConsAlign = "detectTandemCons.align"
        
        groupid = "%s_SATannot_Blaster_%s" %  (self._projectName, typeBlast)
        
        blasterInParallelConfigFileName = "BlasterInParallel.cfg"
        self.writeBlasterInParallelConfig(blasterInParallelConfigFileName)
        
        cmd = "LaunchBlasterInParallel.py"
        cmd += " -q %s" % batchesDir
        cmd += " -p %s" % batchPattern
        cmd += " -C %s" % blasterInParallelConfigFileName
        cmd += " -s %s" % bankName
        cmd += " -o %s" % tandemConsAlign
        cmd += " -u %s" % typeBlast
        cmd += " -S %s" % self._BLR_sensitivity
        cmd += " -g %s" % groupid
        cmd += " -v %s" % (self._verbosity - 1)
        os.system(cmd)
        
        if not os.path.exists( tandemConsAlign ):
            self._logAndRaise("*** Error: %s not in project directory" % ( tandemConsAlign ))
    
        #Coordinate conversion from chunks to chromosomes
        tandemConsAlignOnChr = "detectTandemConsOnChr.align"
        
        iConvCoord = ConvCoord()
        iConvCoord.setConfigFile(self._configFileName)
        iConvCoord.setInputData(tandemConsAlign)
        iConvCoord.setMapData(path2chr2chunkMapFile)
        iConvCoord.setVerbosityLevel(self._verbosity - 2)
        iConvCoord.setOutputData(tandemConsAlignOnChr)
        iConvCoord.run()
        
        if not os.path.exists( tandemConsAlignOnChr ):
            self._logAndRaise("*** Error: %s not in project directory" % ( tandemConsAlignOnChr ))
    
        #Change directory
        os.chdir ( ".." )
        if not os.path.exists( "%s_Matcher" % (prefix) ):
            os.mkdir( "%s_Matcher"% (prefix) )
        else:
            self._log.warn("Warning: directory %s_Matcher already exists" % (prefix))
        os.chdir( "%s_Matcher" % (prefix) )
    
        #Get files to use matcher
        path2tandemConsAlignOnChr = "%s/%s_Pals_%s_%s_%s/%s" % (self._projectDir,self._projectName, self._clustHsp, self._multAlign, "Blaster_Matcher", tandemConsAlignOnChr )
        
        if not os.path.exists( path2tandemConsAlignOnChr  ):
            self._logAndRaise("Error: %s not in project directory" % ( path2tandemConsAlignOnChr ))
        else:
            os.system ("ln -s %s ." % ( path2tandemConsAlignOnChr ) )
    
        # join the matches with Matcher
        groupid = "%s_SATannot_Matcher" % self._projectName
        acronym = "Matcher"
        cDir = os.getcwd()
        if self._tmpDir != "":
            tmpDir = self._tmpDir
        else:
            tmpDir = cDir
            
        iDb = DbFactory.createInstance()
        iTJA = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(iTJA, os.getcwd(), "", "", cDir, tmpDir, "jobs", self._queue, groupid)
        
        lCmdsTuples = []   
        lFiles = FileUtils.getFileNamesList(cDir, tandemConsAlignOnChr)

        for ifile in lFiles:
            lCmdStart = []
            lCmdStart.append("shutil.copy(\"%s/%s\", \".\")" % (cDir, ifile))
            lCmds = []
            lCmds.append(self.getMatcherCmd(iLauncher, ifile))
            lCmdFinish = []
            lCmdFinish.append("if os.path.exists(\"%s.clean_match.map\"):" % ifile)
            lCmdFinish.append("\tshutil.move(\"%s.clean_match.map\", \"%s/.\" )" % (ifile, cDir))
            lCmdFinish.append("if os.path.exists(\"%s.clean_match.path\"):" % ifile)
            lCmdFinish.append("\tshutil.move(\"%s.clean_match.path\", \"%s/.\" )" % (ifile, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)
        
        
        self.removeInPathFileDuplicatedPathOnQueryNameQueryCoordAndSubjectName("detectTandemConsOnChr.align.clean_match.path", "detectTandemConsOnChr.align.clean_match.path.tmp") 

        PathUtils.convertPathFileIntoMapFileWithQueryCoordsOnly("detectTandemConsOnChr.align.clean_match.path.tmp", "detectTandemConsOnChr.align.clean_match.map")

        # clean files
        if self._doClean:
            os.system( "rm -f formatdb.log %s_chunks.fa_cut*" % (self._projectName ) )
            os.system( "rm -f detectTandemConsOnChr.align.clean_match.path.tmp" )
                
        os.chdir ( ".." )
        self._log.info("step 5 finished successfully")


class BuildTandemArrays(PipelineStep):
    def __init__(self, pipelineName, configFileName, configSection, verbosity = 1):
        """
        step 6: finish the construction of the tandem arrays
    
        @param smplAlign: name of the alignment program (Blaster or Pals)
        @type smplAlign: string
    
        @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
        @type clustHsp: string
    
        @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
        @type multAlign: string
        """
      
        PipelineStep.__init__(self, configFileName, configSection, verbosity)
        self._pipelineName = pipelineName
        
    def _setAttributesFromConfig(self):
        iConfigRules = ConfigRules()
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._projectName = iConfig.get("project", "project_name")
        self._projectDir = iConfig.get("project", "project_dir")
        
        self._smplAlign = iConfig.get("self_align", "smplAlign")
        self._clustHsp = iConfig.get("cluster_HSPs", "clustHsp")
        self._multAlign = iConfig.get("build_consensus", "multAlign") 
        
        self._workdir = "%s_%s_%s_%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp, self._multAlign, "Blaster_Matcher","tandemArrays")
        
    def checkConfig(self):
        if os.path.exists(self._workdir):
            self._logAndRaise("Error: directory %s already exists" % self._workdir)

    def run(self):
        self._setAttributesFromConfig()
        if not os.path.exists(self._workdir):
            os.makedirs(self._workdir)
        os.chdir(self._workdir)
        self._log.info("beginning of step 6 self-alignment with PALS clustering with %s multiple alignment with %s mapping consensus with %s finish the detection of the tandem arrays" % (self._clustHsp, self._multAlign, self._smplAlign))
        if not os.path.exists( self._configFileName ):
            os.system ("ln -s ../%s ." % ( self._configFileName ) )
        
        tandemConsAlignOnChr = "detectTandemConsOnChr.align"
    
        inMapFile = "%s.clean_match.map" % ( tandemConsAlignOnChr )
        inMapTable = "%s_%s_map" % (self._projectName, inMapFile.split(".")[0])
        inSetFile = "%s.set" % (inMapFile)
        inSetTable = "%s_%s_set" % (self._projectName, inMapFile.split(".")[0])
        
        path2inMapFile = "%s/%s_Pals_%s_%s_Matcher/%s" % (self._projectDir, self._projectName, self._clustHsp, self._multAlign, inMapFile )
        self._log.info("path2inMapFile : %s" % path2inMapFile)
        if not os.path.exists( path2inMapFile ):
            self._logAndRaise("Error: %s not in project directory" % ( inMapFile ))
        else:
            os.system( "ln -s %s ." % ( path2inMapFile ) )    
    
        #create a table with blaster_matcher results
        self._iDb = DbFactory.createInstance()
        
        self._iDb.createTable(inMapTable, 'map', inMapFile)
        MapUtils.convertMapFileIntoSetFile( inMapFile, inSetFile )
        
        self._iDb.createTable( inSetTable, 'set', inSetFile )
        self._iDb.close()
        
        #finish the tandemearrays detection
        self._log.info("inSetTable :%s" % inSetTable)
        iDetectTandemByCluster = DetectTandemByCluster(tableName=inSetTable, configFileName=self._configFileName, verbosity=self._verbosity-1)
        iDetectTandemByCluster.run() 

        #Creating the reference table on chromosomes
        inFastaFile = "%s.fa" % (self._projectName )
        path2inFastaFile = "%s/%s" % (self._projectDir, inFastaFile )
        self._log.info("path2inFastaFile: %s" % path2inFastaFile)
        if not os.path.exists( path2inFastaFile ):
            self._logAndRaise("Error: %s not in project directory" % (inFastaFile ))
        else:
            os.system( "ln -s %s ." % ( path2inFastaFile ) )
        
        self._iDb = DbFactory.createInstance()
        inSeqFile = "%s" % ( inFastaFile )
        inSeqTable =  "%s_chr_seq" % (self._projectName )
        self._iDb.createTable( inSeqTable,'seq', inSeqFile )
        self._iDb.close()
        
        #Creating the annotation_tables.txt file 
        tableFileName = "annotation_tables.txt"
        tableFile = open( tableFileName, "w" )
        string = "%s_REPET_TAs\tset\t%s_tandem\n" % (self._projectName, inSetTable )
        tableFile.write( string )
        tableFile.close()
        
        #Output as GFF3 file
        #Creating the GFF3 files
        if not os.path.exists( "annotations" ):
            os.mkdir( "annotations" )
        os.chdir( "annotations" )
        inFileName = "../%s" % tableFileName
        iGFF3Maker = GFF3Maker(inFastaName=inSeqTable, tablesFileName=inFileName,verbose=self._verbosity-1)
        iGFF3Maker.run()
        os.system( "cp *.gff3 .." )
        os.chdir( ".." )
        shutil.rmtree( "annotations" )
        os.chdir ( ".." )
        self._log.info("step 6 finished successfully...everything is ready for analysis !")
    
if __name__ == "__main__":
    iSATannotPipeline = SATannotPipeline()
    iSATannotPipeline.setAttributesFromCmdLine()
    iSATannotPipeline.run()      