#!/usr/bin/env python

from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
import sys

"""
This script give statistics for the result of PASTEC.
usage: python StatPastec.py [options]
    --help: this help"
    --i: name of the input file, which is PASTEC type file.(format='OutputFile.classif')"
    --o: name of the output file (default=inFileName+'_statsPASTEC.txt')"
    --v: verbose (default=0/1)"
"""

CLASS_I = 'I'
CLASS_II = 'II'
COLUMNS_NB = 8
CLASS_COLUMN = 4
ORDER_COLUMN = 5
CHIMERIC = " potential chimeric*"
DWICKERCODE = {
                "I" : "RXX", 
                "II" : "DXX", 
                "LTR" : "RLX", 
                "DIRS" : "RYX", 
                "PLE" : "RPX", 
                "LINE" : "RIX", 
                "SINE" : "RSX", 
                "TIR" : "DTX", 
                "Crypton" : "DYX", 
                "Helitron" : "DHX", 
                "Maverick" : "DMX", 
                "MITE" : "DXX-MITE", 
                "LARD" : "RXX-LARD", 
                "TRIM" : "RXX-TRIM"
               }
DORDER2CLASS = {
                    "LTR": CLASS_I,
                    "DIRS": CLASS_I,
                    "PLE": CLASS_I,
                    "LINE": CLASS_I,
                    "SINE": CLASS_I,
                    "LARD": CLASS_I,
                    "TRIM": CLASS_I,
                    "TIR": CLASS_II,
                    "Crypton": CLASS_II,
                    "Maverick": CLASS_II,
                    "Helitron": CLASS_II,
                    "MITE" : CLASS_II
                    }

class StatPastec(object):
    
    def __init__(self, inFileName = "", outFileName = None):
        self._inFileName = inFileName
        self.setOutFileName(outFileName)
        self._verbose = 0
        self._dStat = {}
        self._nbElementsChimeric = 0
        self._nbElementNoCatNoCat = 0
        self._nbElementNoCatSeveralOrders = 0
        self._nbElementSeveralOrders = 0
                
    def setAttributesFromCmdLine(self):
        help = '\
        \nThis script give statistics for the result of PASTEC.\n\n\
        Example 1: python StatPastec.py --infile outfile.classif'
        parser = RepetOptionParser(usage = help, version="StatPastec.py v1.0")
        parser.add_option('-i', '--infile', dest='inFileName', help='name of the input file')
        parser.add_option('-o', '--outfile', dest='outFileName', help='name of the output file', default = None)
        parser.add_option('-v', '--verbosity', dest='verbose', help='verbosity default = 0, else 1', default = 0, type = "int")
        options = parser.parse_args()[0]
        self.setAttributeFromOptions(options)
        
    def setAttributeFromOptions(self, options):
        self.setInputFileName(options.inFileName)
        self.setOutFileName(options.outFileName)
        self.setVerbose(options.verbose)
    
    def setInputFileName(self, inFileName):
        self._inFileName = inFileName
        
    def setOutFileName(self, outFileName):
        if outFileName != None:
            self._outFileName = outFileName
        else:
            self._outFileName = self._inFileName + '_stats.txt'
        
    def setVerbose(self, verbose):
        self._verbose = verbose
         
    def checkOptions(self):
        if self._inFileName == "":
            raise Exception("ERROR: No input file specified!")
        elif not FileUtils.isRessourceExists(self._inFileName):
            raise Exception("ERROR: This input file specified doesn't exist!")
                    
    def run(self):
        if self._verbose > 0:
            print "START %s" % ( sys.argv[0].split("/")[-1] )
            sys.stdout.flush()
            
        self.checkOptions()
        aClassifFile = open(self._inFileName,"r")
        aClassifLine = aClassifFile.readline()
        lineNumber = 0
        totalElement = 0 
        while aClassifLine :
            lineNumber += 1
            lFieldsOfClassifLine = aClassifLine.split("\t")
            if len(lFieldsOfClassifLine) == COLUMNS_NB:
                totalElement += 1 
                self.incrementeForOrder(lFieldsOfClassifLine)
                self.incrementeForSeveralOrders(lFieldsOfClassifLine)
                self.incrementeForNoCat(lFieldsOfClassifLine)
            elif self._verbose > 0:
                print "line %s has not %s columns" % (lineNumber, COLUMNS_NB)
            aClassifLine = aClassifFile.readline()
        self._writeStat(totalElement)

        if self._verbose > 0:
            print "END %s" % ( sys.argv[0].split("/")[-1] )
            sys.stdout.flush()
        
    def incrementeForOrder(self, lFieldsOfClassifLine):
        order = lFieldsOfClassifLine[ORDER_COLUMN]
        if order in DORDER2CLASS.keys():
            self._incrementAStatKeyFound("%s total (%s)" % (order, DWICKERCODE[order]))
            self._incrementAStatKeyFound("Class%s + one order" % DORDER2CLASS[order])
            self._incrementAStatKeyFound("Class%s total (%s)" % (DORDER2CLASS[order], DWICKERCODE[DORDER2CLASS[order]]))
            if "complete" in lFieldsOfClassifLine:
                self._incrementAStatKeyFound("%s comp" % order)
            elif "incomplete" in lFieldsOfClassifLine:
                self._incrementAStatKeyFound("%s incomp" % order)
            if "PotentialChimeric" in lFieldsOfClassifLine:
                self._nbElementsChimeric += 1
                self._incrementAStatKeyFound("Class%s%s" % (DORDER2CLASS[order], CHIMERIC))
                self._incrementAStatKeyFound("".join([order, CHIMERIC]))
        elif order in ["SSR", "rDNA", "PotentialHostGene"]:
            self._incrementAStatKeyFound("%s total" % order)
            if "PotentialChimeric" in lFieldsOfClassifLine:
                self._nbElementsChimeric += 1
                self._incrementAStatKeyFound("".join([order, CHIMERIC]))
        
    def incrementeForNoCat(self, lFieldsOfClassifLine):
        order = str(lFieldsOfClassifLine[ORDER_COLUMN])
        classField = str(lFieldsOfClassifLine[CLASS_COLUMN])
        if order == "noCat" and classField == "noCat":
            self._nbElementNoCatNoCat += 1
        if order == "noCat" and (classField == "I" or classField == "II"):
            self._incrementAStatKeyFound("Class%s + noCat order" % classField)
            self._incrementAStatKeyFound("Class%s total (%s)" % (classField, DWICKERCODE[classField]))
        
    def incrementeForSeveralOrders(self, lFieldsOfClassifLine):
        order = str(lFieldsOfClassifLine[ORDER_COLUMN])
        classField = str(lFieldsOfClassifLine[CLASS_COLUMN])
        if "|" in order:
            self._nbElementsChimeric += 1
            self._nbElementSeveralOrders += 1
            if classField == "noCat":
                self._nbElementNoCatSeveralOrders += 1
            elif classField != "NA":
                self._incrementAStatKeyFound("Class%s + several orders" % classField)
                self._incrementAStatKeyFound("Class%s%s" % (classField, CHIMERIC))
                self._incrementAStatKeyFound("Class%s total (%s)" % (classField, DWICKERCODE[classField]))
                
    def groupList(self, lStatKey):
        lStatKeyClassI = []
        lStatKeyClassII = []
        lStatKeyOrderClassI = []
        lStatKeyOrderClassII = []
        lStatKeyNotAvailable = []
        for aStatKeyName in lStatKey :
            lOrderKeyName = aStatKeyName.split(" ")
            aOrderKeyName = lOrderKeyName[0]
            if DORDER2CLASS.has_key(aOrderKeyName) :
                aClassName = DORDER2CLASS[aOrderKeyName]
                if aClassName == CLASS_I :
                    lStatKeyOrderClassI.append(aStatKeyName)
                elif aClassName == CLASS_II :
                    lStatKeyOrderClassII.append(aStatKeyName)
            elif aOrderKeyName == "ClassI":
                lStatKeyClassI.append(aStatKeyName)
            elif aOrderKeyName == "ClassII":
                lStatKeyClassII.append(aStatKeyName)
            else:
                lStatKeyNotAvailable.append(aStatKeyName)
        lStatKeyClassI.sort()
        lStatKeyClassII.sort()
        lStatKeyOrderClassI.sort()
        lStatKeyOrderClassII.sort()
        lStatKeyNotAvailable.sort() 
        return lStatKeyClassI, lStatKeyClassII, lStatKeyOrderClassI, lStatKeyOrderClassII, lStatKeyNotAvailable
  
    def _incrementAStatKeyFound(self, aKeyName):
        if self._dStat.has_key(aKeyName):
            self._dStat[aKeyName] += 1 
        else:
            self._dStat[aKeyName] = 1

    def _writeStat(self, total):
        outStatFile = open(self._outFileName,"w")
        lStatKey = self._dStat.keys()
        lStatKeyClassI, lStatKeyClassII, lStatKeyOrderClassI, lStatKeyOrderClassII, lStatKeyNotAvailable = self.groupList(lStatKey)
            
        for aStatKeyName in lStatKeyOrderClassI:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
                
        for aStatKeyName in lStatKeyClassI:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
            
        for aStatKeyName in lStatKeyOrderClassII:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
                
        for aStatKeyName in lStatKeyClassII:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
            
        for aStatKeyName in lStatKeyNotAvailable:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
        
        self._writeALineInFile(outStatFile, "noCat class with several orders: %i (%.2f%%)" % (self._nbElementNoCatSeveralOrders, 100*self._nbElementNoCatSeveralOrders/float(total)))
        self._writeALineInFile(outStatFile, "Several orders total: %i (%.2f%%)" % (self._nbElementSeveralOrders, 100*self._nbElementSeveralOrders/float(total)))
        self._writeALineInFile(outStatFile, "Potential chimeric total: %i (%.2f%%)" % (self._nbElementsChimeric, 100*self._nbElementsChimeric/float(total)))
        self._writeALineInFile(outStatFile, "")
        
        self._writeALineInFile(outStatFile, "noCat at class and order levels: %i (%.2f%%)" % (self._nbElementNoCatNoCat, 100*self._nbElementNoCatNoCat/float(total)))
        self._writeALineInFile(outStatFile, "")
            
        self._writeALineInFile(outStatFile, "TOTAL: %i (%.2f%%)" % (total, 100*total/float(total)))
        self._writeALineInFile(outStatFile, "")
        
        self._writeNotes(outStatFile)
        outStatFile.close()

    def _writeNotes(self, outStatFile):
        self._writeALineInFile(outStatFile, "\t----------------------------------------------------------------")
        self._writeALineInFile(outStatFile, "\t-----------------------------NOTES------------------------------")
        self._writeALineInFile(outStatFile, "\t----------------------------------------------------------------")
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "1) Potential chimeric is a tag different from completeness. So in order to have a total of classification type, the number of complete and incomplete are added, but not the number of potential chimeric.")
        self._writeALineInFile(outStatFile, "2) Consensus with several orders are potential chimeric, but no classification is better than the others (same confidence index between all classifications found)")
        self._writeALineInFile(outStatFile, "3) Following all classifications that can be found :")
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "Transposable elements :")
        self._writeALineInFile(outStatFile, "\tClass I")
        for key, value in sorted(DORDER2CLASS.iteritems()):
            if value == CLASS_I:
                self._writeALineInFile(outStatFile, "\t\t%s" % key)
        
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "\tClass II")
        for key, value in sorted(DORDER2CLASS.iteritems()):
            if value == CLASS_II:
                self._writeALineInFile(outStatFile, "\t\t%s" % key)
        
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "Not transposable elements :")
        self._writeALineInFile(outStatFile, "\tPotentialHostGene\n\trDNA\n\tSSR")
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "If a classification is not in the statistical results, it means that none of the analyzed consensus has been classified in this way.")
 
    def _writeALineInFile(self, outStatFile, aLine):
        outStatFile.write(aLine + "\n")
        if self._verbose > 0:
            print (aLine)
                    
if __name__ == '__main__':
    launcher = StatPastec()
    launcher.setAttributesFromCmdLine()
    launcher.run()