#! /usr/bin/env python

import os
import sys
import subprocess

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append(os.environ["REPET_PATH"])
if not "PYTHONPATH" in os.environ.keys():
    os.environ["PYTHONPATH"] = os.environ["REPET_PATH"]
else:
    os.environ["PYTHONPATH"] = "%s:%s" % (os.environ["REPET_PATH"], os.environ["PYTHONPATH"])

from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "repet.launchers"

class LaunchPals(object):
    
    def __init__(self, queryFileName = "", subjectFileName = "", identity = 94, length = 400, doAllByall = False, doClean = False, verbosity = 0):
        self._queryFileName = queryFileName
        self.setSubjectFileName(subjectFileName)
        self._identity = identity
        self._length = length
        self._doAllByall = doAllByall
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        description = "Launch Pals alignement program."
        epilog = "\nExample 1: launch with verbosity , same query and subject (implicit all by all).\n"
        epilog += "\t$ python LaunchPals.py -q file.fa -v 3"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-q", "--query",      dest = "query",         action = "store",       type = "string", help = "query fasta file name [compulsory] [format: fasta]",       default = "")
        parser.add_option("-s", "--subject",    dest = "subject",       action = "store",       type = "string", help = "subject fasta file name [default: query] [format: fasta]", default = "")
        parser.add_option("-d", "--id",         dest = "identity",      action = "store",       type = "int",    help = "Blast identity [default: 90]",                             default = 94)
        parser.add_option("-l", "--length",     dest = "length",        action = "store",       type = "int",    help = "Minimal hit length [default: 100]",                        default = 400)
        parser.add_option("-a", "--aba",        dest = "doAllByall",    action = "store_true",                   help = "all-by-all [default: False]",                        default = False)
        parser.add_option("-c", "--clean",      dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 1)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
    
    def setQueryFileName(self, queryFileName):
        self._queryFileName = queryFileName
    
    def setSubjectFileName(self, subjectFileName):
        if subjectFileName == "":
            self._subjectFileName = self._queryFileName
        else:
            self._subjectFileName = subjectFileName
    
    def setIdentity(self, identity):
        self._identity = identity
        
    def setLength(self, length):
        self._length = length
        
    def setDoAllByall(self, doAllByall):
        self._doAllByall = doAllByall
    
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
    
    def _setAttributesFromOptions(self, options):
        self.setQueryFileName(options.query)
        self.setSubjectFileName(options.subject)
        self.setIdentity(options.identity)
        self.setLength(options.length)
        self.setDoAllByall(options.doAllByall)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)
    
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)  
     
    def _checkOptions(self):
        if self._queryFileName == "":
            self._logAndRaise("ERROR: Missing input fasta file name")
    
    def _getPalsCmd(self):
        lArgs = []
        lArgs.append("-query %s" % self._queryFileName)
        lArgs.append("-target %s" % self._subjectFileName)
        lArgs.append("-length %s" % self._length)
        lArgs.append("-pctid %s" % self._identity)
        lArgs.append("-out %s_%s.gff" % (os.path.basename(os.path.splitext(self._queryFileName)[0]), "Pals"))
        if self._doAllByall:
            lArgs.append("-a")
            
        return self._getSystemCommand("pals", lArgs)
    
    def _getSystemCommand(self, prg, lArgs):
        systemCmd = prg 
        for arg in lArgs:
            systemCmd += " " + arg
        return systemCmd
    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        self._log.info("START LaunchPals")
        
        cmd = self._getPalsCmd()
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        self._log.info("END LaunchPals")

if __name__ == "__main__":
    iLaunchPals = LaunchPals()
    iLaunchPals.setAttributesFromCmdLine()
    iLaunchPals.run()    
