#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.parsing.FastaParser import FastaParser
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.ConfigChecker import ConfigRules
from commons.core.checker.ConfigChecker import ConfigChecker
from commons.core.utils.FileUtils import FileUtils
import shutil
import time
import os
import subprocess

LOG_DEPTH = "repet.tools"
LOG_FORMAT = "%(message)s"

class RemoveRedundancyBasedOnCI(object):

    def __init__(self, inputFastaName = "", inputClassifName = "", configFileName = "", projectName = "removeRedundancy", outFileName = "", identityTh = 95, coverageTh = 98, doClean = False, verbosity = 0):
        self._inputFastaName = inputFastaName
        self._inputClassifName = inputClassifName
        self._configFileName = configFileName
        self._projectName = projectName
        self.setOutFileName(outFileName)
        self._identityThreshold = identityTh
        self._coverageThreshold = coverageTh
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity, LOG_FORMAT)
        self._dHeader2CIAndLength = {}
        self._dHeader2IsKept = {}
        
    def setInClassifFileName(self, inClassifFileName):
        self._inputClassifName = inClassifFileName
        
    def setInFastaFileName(self, inFastaFileName):
        self._inputFastaName = inFastaFileName
        
    def setProjectName(self, projectName):
        self._projectName = projectName
        
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setIdentity(self, identityTh):
        self._identityThreshold = float(identityTh)
        
    def setCoverage(self, coverageTh):
        self._coverageThreshold = float(coverageTh)
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setOutFileName(self, fileName):
        if fileName == "":
            self._outFileName = "%s_withoutRedundancy.fa" % self._projectName
        else:
            self._outFileName = fileName
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def setHeader2IsKeptDict(self, dHeader2IsKept):
        self._dHeader2IsKept = dHeader2IsKept
        
    def setHeader2CIAndLengthDict(self, dHeader2CIAndLength):
        self._dHeader2CIAndLength = dHeader2CIAndLength
        
    def getHeader2IsKeptDict(self):
        return self._dHeader2IsKept
    
    def getHeader2CIAndLengthDict(self):
        return self._dHeader2CIAndLength
    
    def setAttributesFromCmdLine(self):
        description = "Remove redundancy between sequences according to confidence index from PASTEC."
        epilog = "\nExample:\n"
        epilog += "\t$ python RemoveRedundancyBasedOnCI.py -c file.classif -f file.fa --clean -v 3"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-c", "--classif",   dest = "classifFileName",   type = "string", help = "classification file (from PASTEC) [compulsory] [format: classif]")
        parser.add_option("-f", "--fasta",     dest = "fastaFileName",     type = "string", help = "fasta file name [compulsory] [format: fasta]")
        parser.add_option("-C", "--config",    dest = "configFileName",    type = "string", help = "configuration file name (e.g. TEdenovo.cfg)", default = "")
        parser.add_option("-P", "--projectName",dest = "projectName",      type = "string", help = "project name [default: 'removeRedundancy']", default = "removeRedundancy")
        parser.add_option("-I", "--identity",  dest = "identity",          type = "float",  help = "identity threshold [default: 95]", default = 95)
        parser.add_option("-L", "--coverage",  dest = "coverage",          type = "float",  help = "coverage threshold [default: 98]", default = 98)
        parser.add_option("-o", "--output",    dest = "outputFileName",    type = "string", help = "output file name [default: <fasta file name>_withoutRedundancy.fa])", default = "")
        parser.add_option("", "--clean",       dest = "clean",             action="store_true", help = "clean temporary directory [default: False]", default = False)
        parser.add_option("-v", "--verbose",   dest = "verbose",           type = "int",    help = "verbosity [default: 0]", default = 0)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)

    def _setAttributesFromOptions(self, options):
        self.setInClassifFileName(options.classifFileName)
        self.setInFastaFileName(options.fastaFileName)
        self.setConfigFileName(options.configFileName)
        self.setProjectName(options.projectName)
        self.setIdentity(options.identity)
        self.setCoverage(options.coverage)
        self.setOutFileName(options.outputFileName)
        self.setDoClean(options.clean)
        self.setVerbosity(options.verbose)

    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section="project", mandatory=True)
        iConfigRules.addRuleOption(section="project", option ="project_name", mandatory=True, type="string")
        iConfigRules.addRuleSection(section="classif_consensus", mandatory=True)
        iConfigRules.addRuleOption(section="classif_consensus", option ="clean", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section="classif_consensus", option ="min_redundancy_identity", mandatory=True, type="float")
        iConfigRules.addRuleOption(section="classif_consensus", option ="min_redundancy_coverage", mandatory=True, type="float")
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig(iConfig)
        
    def _setAttributesFromConfig(self, iConfig):
        self.setProjectName(iConfig.get("project", "project_name"))
        self.setOutFileName(self._outFileName)
        self.setDoClean(iConfig.get("classif_consensus", "clean"))
        self.setIdentity(iConfig.get("classif_consensus", "min_redundancy_identity"))
        self.setCoverage(iConfig.get("classif_consensus", "min_redundancy_coverage"))
        
    def _checkOptions(self):
        if self._inputFastaName == "" or not os.path.exists(self._inputFastaName):
            self._logAndRaise("ERROR: Missing input fasta file")
        if self._inputClassifName == "" or not os.path.exists(self._inputClassifName):
            self._logAndRaise("ERROR: Missing input classification file")
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)

    def splitFastaFileByChimericTag(self):
        iClassif = open(self._inputClassifName, "r")
        iBioseqDB = BioseqDB(self._inputFastaName)
        dChimTag2FileHandler = {}
        for classifLine in iClassif.readlines():
            lClassifElements = classifLine.split("\t")
            seqName = lClassifElements[0]
            if lClassifElements[3] != "ok":
                key = "potentialChimeric"
            else:
                key = "notPotentialChimeric"
            if not dChimTag2FileHandler.get(key, False):
                dChimTag2FileHandler[key] = open("%s_%s.fa" % (self._projectName, key), "w")  
            iBioseq2 = iBioseqDB.fetch(seqName)       
            iBioseq2.write(dChimTag2FileHandler[key])
        iClassif.close()
        for fileHandler in dChimTag2FileHandler.itervalues():
            fileHandler.close()
    
    def createHeader2CIAndLengthDictFromClassifFile(self):
        with open(self._inputClassifName) as f:
            line = f.readline()
            while line:
                header = line.split()[0]
                length = line.split()[1]
                comments = line.split()[7]
                CI = comments.split("CI=")[1].split(";")[0]
                if CI == 'NA':
                    CI = -1
                self._dHeader2CIAndLength[header] = (int(CI), int(length))
                self._dHeader2IsKept[header] = True
                line = f.readline()
    
    def createFastaAccordingToHeadersDict(self):
        nbRmvSeq = 0
        iFastaParser = FastaParser(self._inputFastaName)
        with open(self._outFileName, "w") as f:
            for iSequence in iFastaParser.getIterator():
                if self._dHeader2IsKept[iSequence.getName()]:
                    f.write(iSequence.printFasta())
                else:
                    nbRmvSeq += 1
        self._log.info("%i sequences removed out of %i" % (nbRmvSeq, iFastaParser.nbReadSequences))
        
    def findSequencesNotKeptFromTabFileAndCIAndLength(self, tabFileName):
        with open(tabFileName) as f:
            f.readline()
            line = f.readline()
            while line:
                identity = float(line.split()[13])
                if identity >= self._identityThreshold:
                    queryName = line.split()[0]
                    subjectName = line.split()[6]
                    queryCov = float(line.split()[4]) * 100
                    subjectCov = float(line.split()[10]) * 100
                    
                    isQueryCovGreaterOrEqualThreshold = (queryCov >= self._coverageThreshold)
                    isQueryCILessThanSubjectCI = (self._dHeader2CIAndLength[queryName][0] < self._dHeader2CIAndLength[subjectName][0])
                    isQueryCIEqualSubjectCI = (self._dHeader2CIAndLength[queryName][0] == self._dHeader2CIAndLength[subjectName][0])
                    isQueryLengthLessOrEqualSubjectLength = (self._dHeader2CIAndLength[queryName][1] <= self._dHeader2CIAndLength[subjectName][1])
                    isQueryLengthGreaterThanSubjectLength = (self._dHeader2CIAndLength[queryName][1] > self._dHeader2CIAndLength[subjectName][1])
                    
                    isSubjectCovGreaterOrEqualThreshold = (subjectCov >= self._coverageThreshold)
                    isSubjectCILessThanQueryCI = (self._dHeader2CIAndLength[subjectName][0] < self._dHeader2CIAndLength[queryName][0])
                    isSubjectCIEqualQueryCI = (self._dHeader2CIAndLength[subjectName][0] == self._dHeader2CIAndLength[queryName][0])
                    isSubjectLengthLessThanQueryLength = (self._dHeader2CIAndLength[subjectName][1] <= self._dHeader2CIAndLength[queryName][1])
                    isSubjectLengthGreaterThanQueryLength = (self._dHeader2CIAndLength[subjectName][1] > self._dHeader2CIAndLength[queryName][1])
                    
                    if isQueryCovGreaterOrEqualThreshold and isQueryCILessThanSubjectCI:
                        self._dHeader2IsKept[queryName] = False
                    elif isQueryCovGreaterOrEqualThreshold and isQueryCIEqualSubjectCI and isQueryLengthLessOrEqualSubjectLength:
                        self._dHeader2IsKept[queryName] = False
                    elif isQueryCovGreaterOrEqualThreshold and isQueryCIEqualSubjectCI and isQueryLengthGreaterThanSubjectLength:
                        self._dHeader2IsKept[subjectName] = False
                    
                    elif isSubjectCovGreaterOrEqualThreshold and isSubjectCILessThanQueryCI:
                        self._dHeader2IsKept[subjectName] = False
                    elif isSubjectCovGreaterOrEqualThreshold and isSubjectCIEqualQueryCI and isSubjectLengthLessThanQueryLength:
                        self._dHeader2IsKept[subjectName] = False
                    elif isSubjectCovGreaterOrEqualThreshold and isSubjectCIEqualQueryCI and isSubjectLengthGreaterThanQueryLength:
                        self._dHeader2IsKept[queryName] = False
                        
                line = f.readline()
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        if self._configFileName:
            self._checkConfig()
        self._checkOptions()
        self._log.info("START RemoveRedundancyBasedOnCI")
        self._log.debug("Fasta file name: %s" % self._inputFastaName)
        self._log.debug("Classif file name: %s" % self._inputClassifName)
        self._log.debug("Project name name: %s" % self._projectName)
        workingDir = self._projectName
        if os.path.exists(workingDir):
            self._logAndRaise("ERROR: %s already exists." % workingDir)
        os.mkdir(workingDir)
        os.chdir(workingDir)
        os.symlink("../%s" % self._inputFastaName, self._inputFastaName)
        os.symlink("../%s" % self._inputClassifName, self._inputClassifName)
        
        self.createHeader2CIAndLengthDictFromClassifFile()
        
        self.splitFastaFileByChimericTag()
        notChimFileName = "%s_notPotentialChimeric.fa" % self._projectName
        chimFileName = "%s_potentialChimeric.fa" % self._projectName
        
        if FileUtils.isRessourceExists(notChimFileName):
            self._runBlasterMatcherAndFindSequencesNotKept(notChimFileName)
        if FileUtils.isRessourceExists(chimFileName):
            self._runBlasterMatcherAndFindSequencesNotKept(chimFileName)
        
        self.createFastaAccordingToHeadersDict()
        
        shutil.move(self._outFileName, "..")
        os.chdir("..")
        if self._doClean:
            self._log.warning("Working directory will be cleaned")
            shutil.rmtree(workingDir)
        self._log.info("END RemoveRedundancyBasedOnCI")

    def _runBlasterMatcherAndFindSequencesNotKept(self, inFileName):
        self._log.info("Blaster (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        outBlasterFileName = "%s.align" % inFileName
        cmd = "blaster"
        cmd += " -q %s" % inFileName
        cmd += " -s %s" % inFileName
        cmd += " -a"
        cmd += " -I %i" % int(self._identityThreshold)
        cmd += " -L 100"
        cmd += " -E 1e-10"
        cmd += " -B %s" % inFileName
        cmd += " 1>&2 >> blasterMatcher.log"
        process = subprocess.Popen(cmd, shell=True)
        self._log.debug("Running : %s" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        self._log.info("Matcher (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        outMatcherFileName = "%s.match.tab" % outBlasterFileName
        cmd = "matcher"
        cmd += " -m %s" % outBlasterFileName
        cmd += " -q %s" % inFileName
        cmd += " -s %s" % inFileName
        cmd += " -j"
        cmd += " -a"
        cmd += " 1>&2 >> blasterMatcher.log"
        process = subprocess.Popen(cmd, shell=True)
        self._log.debug("Running : %s" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        self.findSequencesNotKeptFromTabFileAndCIAndLength(outMatcherFileName)
    
if __name__ == "__main__":
    iRRBOCI = RemoveRedundancyBasedOnCI()
    iRRBOCI.setAttributesFromCmdLine()
    iRRBOCI.run()
