#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.checker.RepetException import RepetException
from commons.core.sql.TableSetAdaptator import TableSetAdaptator
from commons.core.sql.DbMySql import DbMySql
from commons.core.utils.FileUtils import FileUtils
from commons.core.coord.SetUtils import SetUtils
from commons.core.LoggerFactory import LoggerFactory

LOG_DEPTH = "repet.SATannot.DetectTandemByCluster"

#TODO: run without config
class DetectTandemByCluster(object):

    def __init__(self, tableName = "", spacing = 1000,  minSeqPerCluster = 3, configFileName = "SATannot.cfg", verbosity = 0):
        self._tableName = tableName
        self._spacing = spacing
        self._minSeqPerCluster = minSeqPerCluster
        self._configFileName = configFileName
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    ## Set the attributes from the command-line
    #
    def setAttributesFromCmdLine(self):
        description = "Detect tandem arrays from cluster members and store them in a database"
        usage = "DetectTandemByCluster.py -t tableName [options]"

        examples = "\nExample 1: detect tandem repeats with default options and highest verbosity.\n"
        examples += "\t$ python DetectTandemByCluster.py -t inputSetTableName -v 1"
        examples += "\n\t"
        examples += "\t\nExample 2: detect tandem repeats with minimum 10 repeats of the pattern. Each pattern occurrence are distant of 5000 bp maximum.\n"
        examples += "\t$ python DetectTandemByCluster.py -t inputSetTableName -s 5000 -m 10\n"
        
        parser = RepetOptionParser(description = description, usage = usage, version = "v1.0", epilog = examples)
        parser.add_option( "-t", "--table",     dest = "tableName",         type = "string",    help = "set table",                                                             default = "")
        parser.add_option( "-s", "--spacing",   dest = "spacing",           type = "int",       help = "max spacing value (default = 1Kb)",                                     default = 1000)
        parser.add_option( "-m", "--min",       dest = "minSeqPerCluster",  type = "int",       help = "minimum number of sequences per cluster (default=3)",                   default = 3)
        parser.add_option( "-C", "--config",    dest = "configFileName",    type = "string",    help = "configuration file from the SATannot pipeline (default=SATannot.cfg)",  default = "")
        parser.add_option( "-v", "--verbosity",   dest = "verbosity",           type = "int",       help = "verbosity level (default=0/1)",                                         default = 0)

        options, args = parser.parse_args()
        self._setAttributesFromOptions(options)

    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)

    def setTableName(self, tableName):
        self._tableName = tableName
        
    def setSpacing(self, spacing):
        self._spacing = spacing
        
    def setMinSeqPerCluster(self, minSeqPerCluster):
        self._minSeqPerCluster = minSeqPerCluster
        
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setverbosity(self, verbosity):
        self._verbosity = verbosity

    def clusterizeByDistance(self, inLSets):
        lClusterizedSets = []
        lTmp = []
        
        for iset in inLSets:
            if lTmp == []:
                if lClusterizedSets == []:
                    lTmp.append(iset)
                else:
                    if self._mustSetsBeClusterized(lClusterizedSets[-1], iset):
                        self._changeClusterNumber(iset, lClusterizedSets[-1].id)
                        lClusterizedSets.append(iset)
                    else:
                        lTmp.append(iset)
            else:
                if self._mustSetsBeClusterized(lTmp[-1], iset):
                    lTmp.append(iset)
                    if self._haveClusterEnoughMembers(lTmp):
                        self._changeClusterNumber(lTmp, lTmp[0].id)
                        lClusterizedSets.extend(lTmp)
                        lTmp = []
                else:
                    lTmp = []
                    lTmp.append(iset)
        return lClusterizedSets

    def getSetListSortedByPileThenMinThenMax(self, lSets):
        return sorted(lSets, key=lambda iSet: (iSet.getSeqname(), self._getClusterId(iSet.getName()), iSet.getMin(), iSet.getMax()))

    def run(self):
        self._checkAttributes()
        iDb = DbMySql(cfgFileName = self._configFileName)
        tsa = TableSetAdaptator(iDb, self._tableName)
        lofAllSet = tsa.getListOfAllSets()
        lofAllSetMerged = SetUtils.mergeSetsInList(lofAllSet)
        lOfAllSetSorted = self.getSetListSortedByPileThenMinThenMax(lofAllSetMerged)
        
        outTable = "%s_tandem" % self._tableName
        self._log.info("Creation table %s" % (outTable))
        
        tsa = TableSetAdaptator(iDb, outTable)
        
        iDb.createTable(outTable, "set")
        self._log.info("Creation table %s OK" % (outTable))
        self._log.info("Beginning clusterizeByDistance.py ")

        lOfClusterizedSet = self.clusterizeByDistance(lOfAllSetSorted)
        tsa.insertList(lOfClusterizedSet)
        iDb.close()
        self._log.info("End of clusterizeByDistance.py ")
        
        if self._verbosity > 0:
            self._log.info("END %s" % self.__class__.__name__)
            sys.stdout.flush()
        
    def _setAttributesFromOptions(self, options):
        self.setTableName(options.tableName)
        self.setSpacing(options.spacing)
        self.setMinSeqPerCluster(options.minSeqPerCluster)
        self.setConfigFileName(options.configFileName)
        self.setverbosity(options.verbosity)
        
    def _checkAttributes(self):
        if self._tableName == "":
            self._logAndRaise("ERROR: missing input table (-t)")
    
        if self._spacing < 0:
            self._logAndRaise("ERROR: invalid spacing parameter (-s)")
        
        if self._minSeqPerCluster < 0:
            self._logAndRaise("ERROR: invalid minimum sequences per cluster filter (-m)")
        
        if not FileUtils.isRessourceExists(self._configFileName):
            self._logAndRaise("ERROR: configuration file not found (-C)")
            
        if self._verbosity > 0:
            self._log.info("START %s" % self.__class__.__name__)
            sys.stdout.flush()
        
    def _mustSetsBeClusterized(self, set1, set2):
        if self._areSameSeqAndSameCluster(set1, set2):
            if abs(set2.getMin() - set1.getMax()) < self._spacing:
                return True
        return False
        
    def _changeClusterNumber(self, set, clusterNumber):
        if type(set) == type([]):
            for i in range(1, len(set)):
                set[i].id = clusterNumber
        else:
            set.id = clusterNumber
        
    def _areClusterNumberIdentical(self, set1, set2):
        seqName1 = set1.getName()
        seqName2 = set2.getName()
        return self._getClusterId(seqName1) == self._getClusterId(seqName2)
        
    def _areSameSeqAndSameCluster(self, set1, set2):
        return set1.getSeqname() == set2.getSeqname() and self._areClusterNumberIdentical(set1, set2)
    
    def _haveClusterEnoughMembers(self, listMembers):
        return len(listMembers) >= self._minSeqPerCluster
    
    def _getClusterId(self, name):
        return int(name.split("Piler_")[1].split("_")[0])
    
if __name__ == "__main__":
    iDetectTandemByCluster = DetectTandemByCluster()
    iDetectTandemByCluster.setAttributesFromCmdLine()
    iDetectTandemByCluster.run()